import os
import shutil

import click
from edpm.engine.api import EdpmApi
from edpm.engine.output import markup_print as mprint


@click.command("clean", help="Remove installed data for a package from disk (if EDPM owns it).")
@click.argument("dep_name", required=True)
@click.pass_context
def clean_command(ctx, dep_name):
    """
    Removes the installed data for DEP_NAME from disk if EDPM 'owns' it.
    Then clears out the lock file's 'install_path', etc.
    """
    api: EdpmApi = ctx.obj
    api.load_all()  # Ensure plan & lock are loaded

    # 1) Check if the package is installed in the lock
    install_data = api.lock.get_installed_package(dep_name)
    if not install_data:
        mprint("<red>Error:</red> No installation info found for '{}'. Not in lock file.", dep_name)
        raise click.Abort()

    install_path = install_data.get("install_path", "")
    config = install_data.get("built_with_config", {})
    is_owned = install_data.get("owned", True)  # or however you mark "ownership"

    # 2) If no valid install_path or the directory doesn't exist => can't clean
    if not install_path or not os.path.isdir(install_path):
        mprint("<red>Error:</red> '{}' is not currently installed (or directory missing).", dep_name)
        if not is_owned:
            mprint("<yellow>Note:</yellow> '{}' is not owned by EDPM. Remove manually:\n  {}", dep_name, install_path)
        else:
            mprint("<yellow>Nothing to clean on disk for '{}'.</yellow>", dep_name)
        return

    # 3) If EDPM owns it, remove the directories: install_path, build_path, source_path, etc.
    if is_owned:
        dirs_to_remove = [
            config.get("build_path", ""),
            config.get("source_path", "")
        ]
        for path in dirs_to_remove:
            if path and os.path.isdir(path):
                mprint("Removing <magenta>{}</magenta>...", path)
                shutil.rmtree(path, ignore_errors=True)
    else:
        mprint("<yellow>Note:</yellow> '{}' is not owned by EDPM. Remove manually:\n  {}", dep_name, install_path)

    # Rebuild environment scripts
    mprint("\nRebuilding environment scripts...")
    api.save_generator_scripts()

    mprint("<green>Success:</green> Cleaned '{}'", dep_name)



import os
import sys
import subprocess
import click


executed_commands = []

class Command(object):
    def __init__(self):
        pass

    def execute(self):
        raise NotImplementedError()

    def __str__(self):
        return self.__class__.__name__

class RunCommand(Command):
    def __init__(self, args, env_file):
        super(RunCommand, self).__init__()
        self.args = args
        self.env_file = env_file
        self.return_code = None

    def execute(self):
        click.secho("EXECUTING:", fg='blue', bold=True)
        click.echo(self.args)

        # Wrap actual command so we source env_file first in a bash subshell
        if self.env_file:
            shell_cmd = (f'bash -c "source \\"{self.env_file}\\" && '
                         f'echo \'------- env ---------\' && '
                         f'env && '
                         f'echo \'---------------------\' && '
                         f' {self.args}"')
        else:
            shell_cmd = self.args
        self.return_code = subprocess.call(shell_cmd, stdout=sys.stdout, stderr=sys.stderr, shell=True)

        click.secho("Execution done. ", fg='blue', bold=True, nl=False)
        click.echo("Return code = ", nl=False)
        click.secho(f"{self.return_code}", fg='red' if self.return_code else 'green', bold=True)
        click.echo()


class WorkDirCommand(Command):
    def __init__(self, path):
        super(WorkDirCommand, self).__init__()
        self.path = path

    def execute(self):
        click.secho("WORKDIR: ", fg='blue', bold=True, nl=False)
        click.echo(self.path)
        os.chdir(self.path)

class EnvironmentCommand(Command):
    def __init__(self, name, value):
        super(EnvironmentCommand, self).__init__()
        self.name = name
        self.value = value

    def execute(self):
        click.secho("ENV: ", fg='blue', bold=True, nl=False)
        click.echo(f"{self.name} = {self.value}")
        os.environ[self.name] = self.value

def run(args, env_file=None):
    """
    Replaces 'args' with a command that first sources the EDPM environment script if env_file is given,
    ensuring all installed package environment variables are present.
    """

    command = RunCommand(args, env_file)
    _execute_command(command)
    if command.return_code != 0:
        click.secho("ERROR", fg='red', bold=True)
        click.echo(": Command returned nonzero code. The failing command was:")
        click.echo(command.args)
        raise OSError(f"Command failed with return code {command.return_code}")

def workdir(path):
    _execute_command(WorkDirCommand(path))

def env(name, value):
    _execute_command(EnvironmentCommand(name, value))

def _execute_command(command):
    assert isinstance(command, Command), "Passed object is not a Command subclass!"
    command.execute()
    executed_commands.append(command)

def is_not_empty_dir(path):
    return os.path.exists(path) and os.path.isdir(path) and os.listdir(path)



"""
This file provides information of how to build and configure HepMC framework:
https://github.com/AIDASoft/DD4hep
"""

import os
import platform

from edpm.engine.generators.steps import EnvSet, EnvAppend, EnvPrepend, EnvRawText, CmakePrefixPath
from edpm.engine.composed_recipe import ComposedRecipe


class DD4HepRecipe(ComposedRecipe):
    """
    Installs DD4hep from Git + CMake.
    """
    def __init__(self, config):
        self.default_config = {
            'fetch': 'git',
            'make': 'cmake',
            'url': 'https://github.com/AIDASoft/DD4hep.git',
            'branch': 'v01-31'
        }
        super().__init__(name='dd4hep', config=config)

    @staticmethod
    def gen_env(data):
        path = data['install_path']
        bin_path = os.path.join(path, 'bin')
        lib_path = os.path.join(path, 'lib')
        include_path = os.path.join(path, 'include')
        cmake_path = os.path.join(path, 'cmake')

        # EnvPrepend PATH for dd4hep executables
        yield EnvPrepend('PATH', bin_path)

        # Typical dynamic library environment
        yield EnvAppend('LD_LIBRARY_PATH', lib_path)
        if platform.system() == 'Darwin':
            yield EnvAppend('DYLD_LIBRARY_PATH', lib_path)

        # CMake detection
        yield CmakePrefixPath(cmake_path)

        # Root cling might want to find these headers
        yield EnvAppend('ROOT_INCLUDE_PATH', include_path)

        # Optionally source the 'thisdd4hep_only.sh' if it exists
        bash_thisdd4hep_path = os.path.join(bin_path, 'thisdd4hep_only.sh')
        bash_text = (
            f'if [ -f "{bash_thisdd4hep_path}" ] ; then\n'
            f'   source "{bash_thisdd4hep_path}"\n'
            f'fi\n'
        )
        csh_text = (
            f'if ( -f "{bash_thisdd4hep_path}" ) then\n'
            f'    source "{bash_thisdd4hep_path}"\n'
            f'endif\n'
        )
        yield EnvRawText(bash_text, csh_text, None)

        # Also define DD4HEP_DIR
        yield EnvSet('DD4HEP_DIR', path)




"""
https://github.com/JeffersonLab/JANA4ML4FPGA


"""
import os

from edpm.engine.composed_recipe import ComposedRecipe
from edpm.engine.generators.steps import EnvSet, EnvPrepend
from edpm.engine.commands import is_not_empty_dir


class Jana4ml4fpgaRecipe(ComposedRecipe):
    """
    Installs the JANA4ML4FPGA project from Git + CMake.

    Original repo: https://github.com/JeffersonLab/JANA4ML4FPGA
    """

    # OS dependencies (if you want to use `edpm req`):
    os_dependencies = {
        'required': {
            'ubuntu18': "libspdlog-dev",
            'ubuntu22': "libspdlog-dev",
            'centos7':  "spdlog-devel",
            'centos8':  "spdlog-devel"
        },
        'optional': {}
    }

    # Possibly define cmake_deps_flag_names if you use them in your build, e.g. mapping 'root' -> 'ROOT_DIR'
    cmake_deps_flag_names = {
        "root": "ROOT_DIR",
        "jana": "JANA_DIR",
        "genfit": "GENFIT_DIR",
        "eic-smear": "EIC_SMEAR_DIR",
        "hepmc": "HEPMC_DIR"
    }

    def __init__(self, config):
        # Provide minimal defaults; user can override in plan
        self.default_config = {
            'fetch': 'git',
            'make': 'cmake',
            'url': 'https://github.com/JeffersonLab/JANA4ML4FPGA.git',
            'branch': 'main',
            'cxx_standard': 17,
            'cmake_build_type': 'RelWithDebInfo',
            'build_threads': 4,
            # 'cmake_flags': '',     # user can override or set in plan
        }
        super().__init__(name='jana4ml4fpga', config=config)

    def preconfigure(self):
        """
        Combine user-provided cmake flags with defaults for Jana4ml4fpga.
        """
        cxx_std = self.config.get('cxx_standard', 17)
        # We can override the 'source_path' and 'build_threads' as needed
        source_path = self.config.get('source_path', "")
        cmake_build_type = self.config.get('cmake_build_type', 'RelWithDebInfo')
        build_threads = self.config.get('build_threads', 4)

        # Merge user cmake_flags or cmake_custom_flags
        user_flags = self.config.get('cmake_flags', "")
        custom_flags = self.config.get('cmake_custom_flags', "")

        # Construct final cmake invocation
        # The typical pattern is:
        #   cmake -DCMAKE_INSTALL_PREFIX=... -DCMAKE_CXX_STANDARD=... <source> && cmake --build . ...
        # We'll store them in 'cmake_flags' for ComposedRecipe to use
        line = [
            f"-DCMAKE_CXX_STANDARD={cxx_std}",
            f"-DCMAKE_BUILD_TYPE={cmake_build_type}",
            f"-DCMAKE_INSTALL_PREFIX={self.config.get('install_path','')}",
            source_path  # last argument to cmake is the source path
        ]
        merged_flags = " ".join([user_flags, custom_flags, " ".join(line)]).strip()
        self.config['cmake_flags'] = merged_flags
        self.config['build_threads'] = build_threads

    def fetch(self):
        """
        Shallow Git clone if requested; skip if source_path is non-empty.
        """
        source_path = self.config.get('source_path', "")
        if source_path and is_not_empty_dir(source_path):
            return  # Already exists

        from edpm.engine.commands import run
        shallow_flag = ""
        if self.config.get('shallow', False):
            shallow_flag = "--depth 1"
        branch = self.config.get('branch', 'main')
        url = self.config.get('url')
        clone_cmd = f'git clone {shallow_flag} -b {branch} {url} "{source_path}"'
        os.mkdir(source_path, exist_ok=True)
        run(clone_cmd)

    @staticmethod
    def gen_env(data):
        """
        Sets environment variables for JANA4ML4FPGA,
        including PATH, JANA_PLUGIN_PATH, and library paths.
        """
        install_path = data['install_path']
        yield EnvSet('jana4ml4fpga_HOME', install_path)
        yield EnvPrepend('JANA_PLUGIN_PATH', os.path.join(install_path, 'plugins'))
        yield EnvPrepend('PATH', os.path.join(install_path, 'bin'))

        lib_path = os.path.join(install_path, 'lib')
        lib64_path = os.path.join(install_path, 'lib64')
        if os.path.isdir(lib64_path):
            yield EnvPrepend('LD_LIBRARY_PATH', lib64_path)
        else:
            yield EnvPrepend('LD_LIBRARY_PATH', lib_path)

    def patch(self):
        """
        If any patch steps are needed, place them here.
        """
        pass

    def post_install(self):
        """
        Any post-install steps if needed.
        """
        pass





import os
import pytest
from click.testing import CliRunner
import click
from unittest.mock import patch, MagicMock
from ruamel.yaml import YAML

from edpm.cli.init import init_command, get_templates_dir, list_available_templates, load_template_content
from edpm.engine.api import EdpmApi


@click.group()
@click.pass_context
def cli(ctx):
    pass


cli.add_command(init_command, "init")


@pytest.fixture
def runner():
    return CliRunner()


@pytest.fixture
def mock_api(tmp_path):
    """Create a minimal EdpmApi for testing."""
    plan_path = tmp_path / "plan.edpm.yaml"
    lock_path = tmp_path / "plan-lock.edpm.yaml"
    api = EdpmApi(plan_file=str(plan_path), lock_file=str(lock_path))
    return api


@pytest.fixture
def mock_templates_dir(tmp_path):
    """Create a mock templates directory with test templates."""
    templates_dir = tmp_path / "templates"
    templates_dir.mkdir()

    # Create default template
    default_content = """# Default Template
global:
  config:
    cxx_standard: 17

packages:
  - root
"""
    (templates_dir / "default-plan.edpm.yaml").write_text(default_content)

    # Create EIC template
    eic_content = """# EIC Template
global:
  config:
    cxx_standard: 17

packages:
  - root
  - geant4
  - eicrecon
"""
    (templates_dir / "eic-plan.edpm.yaml").write_text(eic_content)

    # Create TDIS template
    tdis_content = """# TDIS Template  
global:
  config:
    cxx_standard: 17

packages:
  - root
  - geant4
  - jana2
"""
    (templates_dir / "tdis-plan.edpm.yaml").write_text(tdis_content)

    return str(templates_dir)


def test_list_templates_command(runner, mock_api, mock_templates_dir):
    """Test the --list-templates option."""
    with patch('edpm.cli.init.get_templates_dir', return_value=mock_templates_dir):
        result = runner.invoke(cli, ["init", "--list-templates"], obj=mock_api)
        assert result.exit_code == 0
        assert "Available templates:" in result.output
        assert "default" in result.output
        assert "eic" in result.output
        assert "tdis" in result.output


def test_init_default_template(runner, mock_api, mock_templates_dir):
    """Test initializing with default template."""
    with patch('edpm.cli.init.get_templates_dir', return_value=mock_templates_dir):
        result = runner.invoke(cli, ["init"], obj=mock_api)
        assert result.exit_code == 0
        assert "Using template: default" in result.output
        assert "Created EDPM plan:" in result.output

        # Check that the file was created with default content
        plan_path = mock_api.plan_file
        assert os.path.exists(plan_path)

        with open(plan_path, 'r') as f:
            content = f.read()
        assert "Default Template" in content
        assert "- root" in content


def test_init_eic_template(runner, mock_api, mock_templates_dir):
    """Test initializing with EIC template."""
    with patch('edpm.cli.init.get_templates_dir', return_value=mock_templates_dir):
        result = runner.invoke(cli, ["init", "-t", "eic"], obj=mock_api)
        assert result.exit_code == 0
        assert "Using template: eic" in result.output
        assert "Created EDPM plan:" in result.output

        # Check that the file was created with EIC content
        plan_path = mock_api.plan_file
        assert os.path.exists(plan_path)

        with open(plan_path, 'r') as f:
            content = f.read()
        assert "EIC Template" in content
        assert "- eicrecon" in content


def test_init_tdis_template(runner, mock_api, mock_templates_dir):
    """Test initializing with TDIS template."""
    with patch('edpm.cli.init.get_templates_dir', return_value=mock_templates_dir):
        result = runner.invoke(cli, ["init", "--template", "tdis"], obj=mock_api)
        assert result.exit_code == 0
        assert "Using template: tdis" in result.output
        assert "Created EDPM plan:" in result.output

        # Check that the file was created with TDIS content
        plan_path = mock_api.plan_file
        assert os.path.exists(plan_path)

        with open(plan_path, 'r') as f:
            content = f.read()
        assert "TDIS Template" in content
        assert "- jana2" in content


def test_init_nonexistent_template(runner, mock_api, mock_templates_dir):
    """Test error when requesting non-existent template."""
    with patch('edpm.cli.init.get_templates_dir', return_value=mock_templates_dir):
        result = runner.invoke(cli, ["init", "-t", "nonexistent"], obj=mock_api)
        assert result.exit_code == 0  # Function doesn't exit with error code, just prints error
        assert "Template 'nonexistent' not found" in result.output
        assert "Available templates: default, eic, tdis" in result.output


def test_init_force_overwrite(runner, mock_api, mock_templates_dir):
    """Test that --force overwrites existing files."""
    plan_path = mock_api.plan_file

    # Create existing file
    with open(plan_path, 'w') as f:
        f.write("existing content")

    with patch('edpm.cli.init.get_templates_dir', return_value=mock_templates_dir):
        result = runner.invoke(cli, ["init", "-t", "eic", "--force"], obj=mock_api)
        assert result.exit_code == 0
        assert "Using template: eic" in result.output
        assert "Created EDPM plan:" in result.output

        # Check that the file was overwritten
        with open(plan_path, 'r') as f:
            content = f.read()
        assert "existing content" not in content
        assert "EIC Template" in content


def test_init_no_force_existing_file(runner, mock_api, mock_templates_dir):
    """Test that existing files are not overwritten without --force."""
    plan_path = mock_api.plan_file

    # Create existing file
    with open(plan_path, 'w') as f:
        f.write("existing content")

    with patch('edpm.cli.init.get_templates_dir', return_value=mock_templates_dir):
        result = runner.invoke(cli, ["init", "-t", "eic"], obj=mock_api)
        assert result.exit_code == 0
        assert "already exists" in result.output
        assert "Use --force to overwrite" in result.output

        # Check that the file was NOT overwritten
        with open(plan_path, 'r') as f:
            content = f.read()
        assert content == "existing content"


def test_templates_directory_functions(mock_templates_dir):
    """Test the helper functions for template management."""
    with patch('edpm.cli.init.get_templates_dir', return_value=mock_templates_dir):
        # Test list_available_templates
        templates = list_available_templates()
        assert templates == ["default", "eic", "tdis"]

        # Test load_template_content
        default_content = load_template_content("default")
        assert "Default Template" in default_content

        eic_content = load_template_content("eic")
        assert "EIC Template" in eic_content

        # Test loading non-existent template
        with pytest.raises(FileNotFoundError):
            load_template_content("nonexistent")


def test_no_templates_directory(runner, mock_api):
    """Test behavior when templates directory doesn't exist."""
    with patch('edpm.cli.init.get_templates_dir', return_value=None):
        # Test list templates
        result = runner.invoke(cli, ["init", "--list-templates"], obj=mock_api)
        assert result.exit_code == 0
        assert "No templates found" in result.output

        # Test init with template
        result = runner.invoke(cli, ["init", "-t", "eic"], obj=mock_api)
        assert result.exit_code == 0
        assert "Templates directory not found" in result.output


from edpm.engine.composed_recipe import ComposedRecipe
from edpm.engine.config import ConfigNamespace

class GitCmakeRecipe(ComposedRecipe):
    """
    A refactored Git+CMake recipe that uses the new EDPM pipeline:
      fetch -> patch -> build -> install -> post_install,
    plus a gen_env() returning environment actions.

    This version references self.config[...] everywhere, so the user
    clearly sees all settings (like repo_address, branch, etc.) in config.
    """

    def __init__(self, name: str, config: ConfigNamespace = None):
        super().__init__(name, config)

        # Provide some defaults if they aren't set in the plan file
        self.config['fetch'] = 'git'
        self.config['make'] =  'cmake'

    def preconfigure(self):
        super().preconfigure()




import os
import tempfile
import pytest
from unittest.mock import patch, MagicMock
from ruamel.yaml import YAML

from edpm.engine.api import EdpmApi
from edpm.engine.recipe import Recipe
from edpm.engine.config import ConfigNamespace

class TestRecipe(Recipe):
    """A test recipe that tracks method calls without doing real work."""

    def __init__(self, config=None):
        config = config or ConfigNamespace()
        super().__init__(config)
        self.calls = []

    def preconfigure(self):
        self.calls.append("preconfigure")
        # Set paths that would normally be set by use_common_dirs_scheme
        app_path = str(self.config.get("app_path", "/tmp"))
        self.config["install_path"] = os.path.join(app_path, "test-install")
        self.config["source_path"] = os.path.join(app_path, "test-source")
        self.config["build_path"] = os.path.join(app_path, "test-build")

    def fetch(self):
        self.calls.append("fetch")

    def patch(self):
        self.calls.append("patch")

    def build(self):
        self.calls.append("build")

    def install(self):
        self.calls.append("install")

    def post_install(self):
        self.calls.append("post_install")


def test_install_dependency():
    """Test that installing a dependency works correctly."""

    with tempfile.TemporaryDirectory() as tmpdir:
        # Setup plan and lock files
        plan_path = os.path.join(tmpdir, "plan.edpm.yaml")
        lock_path = os.path.join(tmpdir, "lock.edpm.yaml")

        # Create minimal plan file with a test package
        plan_data = {
            "global": {"config": {}},
            "packages": ["test"]
        }
        yaml = YAML()
        with open(plan_path, "w", encoding="utf-8") as f:
            yaml.dump(plan_data, f)

        # Create minimal lock file
        lock_data = {
            "file_version": 1,
            "top_dir": tmpdir,
            "packages": {}
        }
        with open(lock_path, "w", encoding="utf-8") as f:
            yaml.dump(lock_data, f)

        # Initialize API
        api = EdpmApi(plan_file=plan_path, lock_file=lock_path)

        # Create test recipe
        test_recipe = TestRecipe(ConfigNamespace(app_path=tmpdir))

        # Patch external interactions
        with patch("edpm.engine.commands.run"), \
                patch("edpm.engine.commands.workdir"), \
                patch("edpm.engine.generators.environment_generator.EnvironmentGenerator.save_environment_with_infile"), \
                patch.object(api.recipe_manager, "create_recipe") as mock_create_recipe:

            # Make create_recipe return our test recipe
            mock_create_recipe.return_value = test_recipe

            # Load plan and lock
            api.load_all()

            # Run installation
            api.install_dependency_chain(dep_names=["test"], force=False)

            # Verify recipe methods were called in order
            assert test_recipe.calls == [
                "preconfigure", "fetch", "patch", "build", "install", "post_install"
            ]

            # Verify lock file was updated
            api.lock.load(lock_path)  # Reload to see saved changes
            assert "test" in api.lock.get_installed_packages()
            package_info = api.lock.get_installed_package("test")
            assert "install_path" in package_info
            assert package_info["install_path"] == os.path.join(tmpdir, "test-install")
            assert "built_with_config" in package_info


def test_install_already_installed():
    """Test installing a package that's already installed."""

    with tempfile.TemporaryDirectory() as tmpdir:
        # Setup plan and lock files
        plan_path = os.path.join(tmpdir, "plan.edpm.yaml")
        lock_path = os.path.join(tmpdir, "lock.edpm.yaml")

        # Create plan file
        plan_data = {
            "global": {"config": {}},
            "packages": ["test"]
        }
        yaml = YAML()
        with open(plan_path, "w", encoding="utf-8") as f:
            yaml.dump(plan_data, f)

        # Create lock file with test package already installed
        install_path = os.path.join(tmpdir, "existing-install")
        os.makedirs(install_path, exist_ok=True)  # Create the directory so it exists

        lock_data = {
            "file_version": 1,
            "top_dir": tmpdir,
            "packages": {
                "test": {
                    "install_path": install_path,
                    "built_with_config": {"some_key": "some_value"}
                }
            }
        }
        with open(lock_path, "w", encoding="utf-8") as f:
            yaml.dump(lock_data, f)

        # Initialize API
        api = EdpmApi(plan_file=plan_path, lock_file=lock_path)

        # Create test recipe
        test_recipe = TestRecipe(ConfigNamespace(app_path=tmpdir))

        # Patch external interactions
        with patch("edpm.engine.commands.run"), \
                patch("edpm.engine.commands.workdir"), \
                patch("edpm.engine.generators.environment_generator.EnvironmentGenerator.save_environment_with_infile"), \
                patch.object(api.recipe_manager, "create_recipe") as mock_create_recipe:

            # Setup the mock
            mock_create_recipe.return_value = test_recipe

            # Load plan and lock
            api.load_all()

            # Run installation - shouldn't actually do anything since it's already installed
            api.install_dependency_chain(dep_names=["test"], force=False)

            # Verify recipe was not used (no methods called)
            assert not test_recipe.calls

            # Verify lock file still has the original data
            api.lock.load(lock_path)
            assert "test" in api.lock.get_installed_packages()
            package_info = api.lock.get_installed_package("test")
            assert package_info["install_path"] == install_path
            assert package_info["built_with_config"] == {"some_key": "some_value"}


def test_install_with_force():
    """Test that force=True reinstalls even if already installed."""

    with tempfile.TemporaryDirectory() as tmpdir:
        # Setup plan and lock files
        plan_path = os.path.join(tmpdir, "plan.edpm.yaml")
        lock_path = os.path.join(tmpdir, "lock.edpm.yaml")

        # Create plan file
        plan_data = {
            "global": {"config": {}},
            "packages": ["test"]
        }
        yaml = YAML()
        with open(plan_path, "w", encoding="utf-8") as f:
            yaml.dump(plan_data, f)

        # Create lock file with test package already installed
        install_path = os.path.join(tmpdir, "existing-install")
        os.makedirs(install_path, exist_ok=True)

        lock_data = {
            "file_version": 1,
            "top_dir": tmpdir,
            "packages": {
                "test": {
                    "install_path": install_path,
                    "built_with_config": {"some_key": "original_value"}
                }
            }
        }
        with open(lock_path, "w", encoding="utf-8") as f:
            yaml.dump(lock_data, f)

        # Initialize API
        api = EdpmApi(plan_file=plan_path, lock_file=lock_path)

        # Create test recipe
        test_recipe = TestRecipe(ConfigNamespace(app_path=tmpdir))

        # Patch external interactions
        with patch("edpm.engine.commands.run"), \
                patch("edpm.engine.commands.workdir"), \
                patch("edpm.engine.generators.environment_generator.EnvironmentGenerator.save_environment_with_infile"), \
                patch.object(api.recipe_manager, "create_recipe") as mock_create_recipe:

            # Setup mock to return our test recipe
            mock_create_recipe.return_value = test_recipe

            # Load plan and lock
            api.load_all()

            # Run installation with force=True to trigger reinstall
            api.install_dependency_chain(dep_names=["test"], force=True)

            # Verify recipe methods were called (force reinstall should use the recipe)
            assert test_recipe.calls == [
                "preconfigure", "fetch", "patch", "build", "install", "post_install"
            ]

            # Verify lock file was updated with new paths
            api.lock.load(lock_path)
            assert "test" in api.lock.get_installed_packages()
            package_info = api.lock.get_installed_package("test")
            assert package_info["install_path"] == os.path.join(tmpdir, "test-install")
            assert "built_with_config" in package_info
            # Original values should be replaced
            assert package_info["built_with_config"].get("some_key") != "original_value"


def test_install_existing_package():
    """Test installing a package with the 'existing' field."""

    with tempfile.TemporaryDirectory() as tmpdir:
        # Setup plan and lock files
        plan_path = os.path.join(tmpdir, "plan.edpm.yaml")
        lock_path = os.path.join(tmpdir, "lock.edpm.yaml")

        # Create minimal plan file with a test package with 'existing' field
        plan_data = {
            "global": {"config": {}},
            "packages": [{"test": {"existing": os.path.join(tmpdir, "external-install")}}]
        }
        yaml = YAML()
        with open(plan_path, "w", encoding="utf-8") as f:
            yaml.dump(plan_data, f)

        # Create minimal lock file
        lock_data = {
            "file_version": 1,
            "top_dir": tmpdir,
            "packages": {}
        }
        with open(lock_path, "w", encoding="utf-8") as f:
            yaml.dump(lock_data, f)

        # Initialize API
        api = EdpmApi(plan_file=plan_path, lock_file=lock_path)

        # Create test recipe
        test_recipe = TestRecipe(ConfigNamespace(app_path=tmpdir))

        # Patch external interactions
        with patch("edpm.engine.commands.run"), \
                patch("edpm.engine.commands.workdir"), \
                patch("edpm.engine.generators.environment_generator.EnvironmentGenerator.save_environment_with_infile"), \
                patch.object(api.recipe_manager, "create_recipe") as mock_create_recipe:

            # Setup the mock
            mock_create_recipe.return_value = test_recipe

            # Load plan and lock
            api.load_all()

            # Run installation
            api.install_dependency_chain(dep_names=["test"], force=False)

            # Verify recipe was not used (no methods called)
            assert not test_recipe.calls  # Should not call recipe methods

            # Verify lock file has expected data
            api.lock.load(lock_path)
            assert "test" in api.lock.get_installed_packages()
            package_info = api.lock.get_installed_package("test")
            assert package_info["install_path"] == os.path.join(tmpdir, "external-install")
            assert package_info["owned"] is False  # Should be marked as not owned



"""
This file provides information of how to build and configure CERN.ROOT:
https://github.com/root-project/root

"""

import os
import platform

from edpm.engine.composed_recipe import ComposedRecipe
from edpm.engine.generators.steps import EnvSet, EnvPrepend, EnvRawText, CmakePrefixPath


ROOTSYS = "ROOTSYS"

class RootRecipe(ComposedRecipe):
    """
    Installs CERN ROOT from Git + CMake, retaining advanced logic:
    - Minimum cxx_standard=14
    - Python detection
    - root7=ON, gdml=ON, etc.
    - 'thisroot.sh' environment script
    - Optional conda skip
    """

    # OS dependencies (optional). You can leave this out if you don't use `edpm req`.
    os_dependencies = {
        'required': {
            'ubuntu18': "dpkg-dev binutils libx11-dev libxpm-dev libxft-dev libxext-dev liblzma-dev",
            'ubuntu22': "dpkg-dev binutils libx11-dev libxpm-dev libxft-dev libxext-dev liblzma-dev",
            'centos7':  "gcc binutils libX11-devel libXpm-devel libXft-devel libXext-devel",
            'centos8':  "gcc binutils libX11-devel libXpm-devel libXft-devel libXext-devel"
        },
        'optional': {
            'ubuntu18': "gfortran libssl-dev libpcre3-dev "
                        "xlibmesa-glu-dev libglew-dev libftgl-dev "
                        "libmysqlclient-dev libfftw3-dev libcfitsio-dev "
                        "graphviz-dev libavahi-compat-libdnssd-dev "
                        "libldap2-dev python3-dev libxml2-dev libkrb5-dev "
                        "libgsl0-dev",
            'ubuntu22': "gfortran libssl-dev libpcre3-dev "
                        "xlibmesa-glu-dev libglew-dev libftgl-dev "
                        "libmysqlclient-dev libfftw3-dev libcfitsio-dev "
                        "graphviz-dev libavahi-compat-libdnssd-dev "
                        "libldap2-dev python3-dev libxml2-dev libkrb5-dev "
                        "libgsl0-dev",
            'centos7':  "gcc-gfortran openssl-devel pcre-devel "
                        "mesa-libGL-devel mesa-libGLU-devel glew-devel ftgl-devel mysql-devel "
                        "fftw-devel cfitsio-devel graphviz-devel "
                        "avahi-compat-libdns_sd-devel libldap-dev python-devel "
                        "libxml2-devel gsl-static",
            'centos8':  "gcc-gfortran openssl-devel pcre-devel "
                        "mesa-libGL-devel mesa-libGLU-devel ftgl-devel mysql-devel "
                        "fftw-devel cfitsio-devel graphviz "
                        "openldap-devel python3-devel "
                        "libxml2-devel gsl-devel"
        },
    }

    def __init__(self, config):
        # Provide minimal defaults. Many can be overridden from plan.
        self.default_config = {
            'fetch': 'git',
            'make': 'cmake',
            'url': 'https://github.com/root-project/root.git',  # main GitHub
            'branch': 'v6-36-00',
            'shallow': True,  # means use `--depth 1`
            'cxx_standard': '17',
            'cmake_build_type': 'RelWithDebInfo',
            # If user doesn't override, we'll set them in preconfigure().
            'cmake_flags': "-Wno-dev",
                "-Droot7=ON "
                "-Dgdml=ON "
                "-Dxrootd=OFF "
                "-Dmysql=OFF "
                "-Dpythia6=OFF "
                "-Dpythia6_nolink=OFF "
                "-Dpythia8=OFF "
                "-Dhttp=ON "
            'build_threads': 4,
        }
        super().__init__(name='root', config=config)


    @staticmethod
    def gen_env(data):
        """
        Replicates environment logic from old recipe (including conda skip).
        """
        install_path = data['install_path']
        bin_path = os.path.join(install_path, 'bin')
        lib_path = os.path.join(install_path, 'lib')
        cmake_path = os.path.join(install_path, 'cmake')
        yield CmakePrefixPath(cmake_path)

        # We'll skip calling 'thisroot' if conda is found
        is_under_conda = os.environ.get('ROOT_INSTALLED_BY_CONDA', False)

        def update_python_environment():
            """
            This function is called in-process for Python environment updates.
            We set environment variables like PATH, LD_LIBRARY_PATH, etc.
            """
            # You can replicate old logic
            yield EnvSet('ROOTSYS', install_path)
            yield EnvPrepend('PATH', bin_path)
            yield EnvPrepend('LD_LIBRARY_PATH', lib_path)
            if platform.system() == 'Darwin':
                yield EnvPrepend('DYLD_LIBRARY_PATH', lib_path)
            # Put the .so python modules into PYTHONPATH
            yield EnvPrepend('PYTHONPATH', lib_path)
            # Possibly also add a jupyter path
            jup_path = os.path.join(install_path, 'etc', 'notebook')
            yield EnvPrepend('JUPYTER_PATH', jup_path)
            # EnvPrepend overall prefix
            yield EnvPrepend('CMAKE_PREFIX_PATH', install_path)

        # Build the scripts to source
        bash_thisroot = os.path.join(bin_path, 'thisroot.sh')
        csh_thisroot = os.path.join(bin_path, 'thisroot.csh')

        # If user is not under conda, we source them
        if not is_under_conda:
            bash_text = (
                f'\nif [ -f "{bash_thisroot}" ]; then\n'
                f'   source "{bash_thisroot}"\n'
                f'fi\n'
            )
            csh_text = (
                f'\nif ( -f "{csh_thisroot}" ) then\n'
                f'   source "{csh_thisroot}"\n'
                f'endif\n'
            )
        else:
            # Under conda? skip
            bash_text = "# Skipping thisroot.sh under conda"
            csh_text = "# Skipping thisroot.csh under conda"

        # Provide the Python in-process environment update function
        def python_env_updater():
            for action in update_python_environment():
                action.update_python_env()

        yield EnvRawText(bash_text, csh_text, python_env_updater)


# Optional utility function from old code if you want it:
def root_find():
    """
    Looks for CERN ROOT by checking $ROOTSYS.
    Returns a list with one element if found, or empty if not found.
    """
    if ROOTSYS not in os.environ:
        print("<red>ROOTSYS</red> not found in the environment")
        return []
    root_sys_path = os.environ[ROOTSYS]
    if not os.path.isdir(root_sys_path):
        print("WARNING: ROOTSYS points to nonexistent directory:", root_sys_path)
        return []
    return [root_sys_path]




import os
import pytest

from edpm.engine.api import EdpmApi
from edpm.engine.generators.environment_generator import EnvironmentGenerator
from edpm.engine.generators.cmake_generator import CmakeGenerator
from edpm.engine.planfile import PlanFile

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
TEST_TMP_DIR = os.path.join(TEST_DIR, "tmp")

BASH_IN_PATH = os.path.join(TEST_TMP_DIR, "env_bash_in.sh")
BASH_OUT_PATH = os.path.join(TEST_TMP_DIR, "env_bash_out.sh")
CSH_IN_PATH = os.path.join(TEST_TMP_DIR, "env_csh_in.csh")
CSH_OUT_PATH = os.path.join(TEST_TMP_DIR, "env_csh_out.csh")


@pytest.fixture
def setup_test_environment():
    """Sets up a temporary directory and sample 'in' files with placeholders."""
    os.makedirs(TEST_TMP_DIR, exist_ok=True)

    # Create an initial env_bash_in file with a placeholder
    with open(BASH_IN_PATH, "w") as f:
        f.write("# Initial content (bash)\n# {{{EDPM-CONTENT}}}\n# End content\n")

    # Create an initial env_csh_in file with a placeholder
    with open(CSH_IN_PATH, "w") as f:
        f.write("# Initial content (csh)\n# {{{EDPM-CONTENT}}}\n# End content\n")

    yield  # Tests run

    # Cleanup
    for path in [BASH_IN_PATH, BASH_OUT_PATH, CSH_IN_PATH, CSH_OUT_PATH]:
        if os.path.exists(path):
            os.remove(path)
    try:
        os.rmdir(TEST_TMP_DIR)
    except OSError:
        pass


def _make_minimal_api():
    """Returns an EdpmApi with a minimal in-memory plan + lock, no actual disk files."""
    api = EdpmApi(plan_file="in-memory-plan.yaml", lock_file="in-memory-lock.yaml")
    raw_data = {
        "global": {
            "config": {},
            "environment": [
                {"set": {"MY_VAR": "some_value"}}
            ]
        },
        "packages": []
    }
    from edpm.engine.planfile import PlanFile
    api.plan = PlanFile(raw_data)
    api.lock.data = {
        "top_dir": "/fake/top_dir",
        "packages": {}
    }
    return api


def test_env_save_with_in_out_files(setup_test_environment):
    """
    If 'in' files exist with the EDPM placeholder, verify merging environment lines.
    """
    api = _make_minimal_api()
    api.plan.data["global"]["config"].update({
        "env_bash_in": BASH_IN_PATH,
        "env_bash_out": BASH_OUT_PATH,
        "env_csh_in": CSH_IN_PATH,
        "env_csh_out": CSH_OUT_PATH
    })

    env_gen = EnvironmentGenerator(plan=api.plan, lock=api.lock, recipe_manager=api.recipe_manager)
    # Merge for bash
    env_gen.save_environment_with_infile("bash", BASH_IN_PATH, BASH_OUT_PATH)
    # Merge for csh
    env_gen.save_environment_with_infile("csh", CSH_IN_PATH, CSH_OUT_PATH)

    with open(BASH_OUT_PATH, "r") as f:
        content = f.read()
        assert "# Initial content (bash)" in content
        assert "export MY_VAR=\"some_value\"" in content

    with open(CSH_OUT_PATH, "r") as f:
        content = f.read()
        assert "# Initial content (csh)" in content
        assert "setenv MY_VAR \"some_value\"" in content


def test_env_save_without_in_files(setup_test_environment):
    """
    If no 'in' files are defined, we want to generate environment scripts from scratch.
    We'll pass in None (instead of empty string) to skip merging logic.
    """
    api = _make_minimal_api()
    env_gen = EnvironmentGenerator(plan=api.plan, lock=api.lock, recipe_manager=api.recipe_manager)

    # Instead of "", pass None to avoid FileNotFoundError
    env_gen.save_environment_with_infile("bash", None, BASH_OUT_PATH)
    env_gen.save_environment_with_infile("csh", None, CSH_OUT_PATH)

    with open(BASH_OUT_PATH) as f:
        content = f.read()
        assert "#!/usr/bin/env bash" in content
        assert "export MY_VAR=\"some_value\"" in content

    with open(CSH_OUT_PATH) as f:
        content = f.read()
        assert "#!/usr/bin/env csh" in content
        assert "setenv MY_VAR \"some_value\"" in content


def test_cmake_toolchain_in_out(setup_test_environment):
    """
    1) Create minimal plan, set cmake_toolchain_in/out in global config.
    2) Save toolchain with placeholders.
    3) Confirm final file merges + includes EDPM's 'Automatically generated' lines.
    """
    cm_in = os.path.join(TEST_TMP_DIR, "cmake_toolchain_in.cmake")
    cm_out = os.path.join(TEST_TMP_DIR, "cmake_toolchain_out.cmake")
    with open(cm_in, "w") as f:
        f.write("# Initial CMake content\n# {{{EDPM-CONTENT}}}\n")

    api = _make_minimal_api()
    api.plan.data["global"]["config"].update({
        "cmake_toolchain_in": cm_in,
        "cmake_toolchain_out": cm_out
    })

    cm_gen = CmakeGenerator(plan=api.plan, lock=api.lock, recipe_manager=api.recipe_manager)
    cm_gen.save_toolchain_with_infile(cm_in, cm_out)

    with open(cm_out) as f:
        content = f.read()
        assert "# Initial CMake content" in content
        # The new generator at least writes a comment header
        assert "# Automatically generated by EDPM" in content


def test_cmake_presets_in_out(setup_test_environment):
    """
    1) Minimal plan, set cmake_presets_in/out
    2) Save presets
    3) Confirm final JSON merges
    """
    import json

    cm_in = os.path.join(TEST_TMP_DIR, "cmake_presets_in.json")
    cm_out = os.path.join(TEST_TMP_DIR, "cmake_presets_out.json")
    initial_json = '{"version": 3, "configurePresets": [{"name":"base"}]}'
    with open(cm_in, "w") as f:
        f.write(initial_json)

    api = _make_minimal_api()
    api.plan.data["global"]["config"].update({
        "cmake_presets_in": cm_in,
        "cmake_presets_out": cm_out
    })

    cm_gen = CmakeGenerator(plan=api.plan, lock=api.lock, recipe_manager=api.recipe_manager)
    cm_gen.save_presets_with_infile(cm_in, cm_out)

    # Quick substring check is still fine:
    with open(cm_out) as f:
        content = f.read()
        # We still confirm "version": 3 is there
        assert '"version": 3' in content

    # Now parse the JSON properly:
    with open(cm_out, "r") as f:
        data = json.load(f)

    # We expect 'version' = 3
    assert data["version"] == 3

    # Confirm that "base" is among the preset names
    presets = data.get("configurePresets", [])
    preset_names = [p.get("name", "") for p in presets]
    assert "base" in preset_names, f"Expected 'base' in {preset_names}"

    # If you want to ensure the generator added an “edpm” preset:
    # assert "edpm" in preset_names




"""
EPIC DD4Hep geometry repository
"""
import os

from edpm.engine.generators.steps import EnvSet, EnvPrepend
from edpm.engine.composed_recipe import ComposedRecipe


class EpicRecipe(ComposedRecipe):
    """
    Installs the ePIC detector software from Git + CMake.
    """
    def __init__(self, config):
        self.default_config = {
            'fetch': 'git',
            'make': 'cmake',
            'url': 'https://github.com/eic/epic.git',
            'branch': '25.02.0'
        }
        super().__init__(name='epic', config=config)

    @staticmethod
    def gen_env(data):
        path = data['install_path']

        yield EnvPrepend('LD_LIBRARY_PATH', os.path.join(path, 'lib'))
        yield EnvPrepend('PATH', os.path.join(path, 'bin'))

        yield EnvSet('DETECTOR_PATH', os.path.join(path, 'share', 'epic'))
        yield EnvSet('BEAMLINE', 'epic')
        yield EnvSet('BEAMLINE_PATH', os.path.join(path, 'share', 'epic'))
        yield EnvSet('BEAMLINE_CONFIG', 'epic')




import click
import os
import shutil

from edpm.engine.api import EdpmApi
from edpm.engine.output import markup_print as mprint


_help_option_lock = "Removes only lock file record without touching installation"
_help_option_all = "Removes lock record and package folder from disk"
_help_option_auto = "Removes from lock and disk if(!) the package is owned by edpm"


@click.command("rm")
@click.argument('package_name', nargs=1, metavar='<package-name>')
@click.option('--lock', 'mode', flag_value='lock', help=_help_option_lock)
@click.option('--all', 'mode', flag_value='all', help=_help_option_all)
@click.option('--auto', 'mode', flag_value='auto', help=_help_option_auto, default=True)
@click.pass_context
def rm_command(ctx, package_name, mode):
    """Removes a package.
    By default, deletes record from edpm lock file and the disk folders if the package is 'owned' by edpm.

    Usage:
        edpm rm <package-name>         # removes the package
        edpm rm <package-name> --lock  # removes only from lock file
        edpm rm <package-name> --all   # forces removal of all files even for non-owned packages

    """
    api = ctx.obj
    assert isinstance(api, EdpmApi)

    # Ensure lock file exists
    api.ensure_lock_exists()

    # Load plan and lock if not already loaded
    if not api.plan or not api.lock:
        api.load_all()

    # Check if package exists
    if not api.plan.find_package(package_name):
        mprint("<red>Error:</red> Package '{}' not found in the plan.", package_name)
        raise click.Abort()

    # Check if package is installed
    if not api.lock.is_installed(package_name):
        mprint("<red>Error:</red> Package '{}' is not installed.", package_name)
        raise click.Abort()

    # Get package data
    package_data = api.lock.get_installed_package(package_name)
    install_path = package_data.get("install_path", "")

    mprint("<blue><b>Removing package: </b></blue> {}", package_name)
    mprint("<blue><b>Installation path: </b></blue> {}\n", install_path)

    # Determine if we should remove folders based on mode and ownership
    remove_folders = False
    if mode == 'all':
        remove_folders = True
    elif mode == 'auto':
        remove_folders = package_data.get("owned", True)
        if not remove_folders:
            mprint("<b>(!)</b> Package is not 'owned' by edpm. The record is removed from lock file but\n"
                   "<b>(!)</b> you have to remove the folder manually:\n{}\n", install_path)

    # Remove package from lock file
    api.lock.remove_package(package_name)
    api.lock.save()

    # Update environment scripts
    mprint("Updating environment script files...\n")
    api.save_generator_scripts()

    # Remove folders if needed
    if remove_folders and os.path.exists(install_path):
        mprint("Removing installation folder from disk...\n")
        try:
            shutil.rmtree(install_path)
            mprint("<green>Successfully removed folder:</green> {}", install_path)
        except Exception as e:
            mprint("<red>Error removing folder:</red> {}", str(e))

        # Remove source and build folders if they exist
        source_path = package_data.get("source_path", "")
        if source_path and os.path.exists(source_path):
            try:
                shutil.rmtree(source_path)
                mprint("<green>Successfully removed source folder:</green> {}", source_path)
            except Exception as e:
                mprint("<red>Error removing source folder:</red> {}", str(e))

        build_path = package_data.get("build_path", "")
        if build_path and os.path.exists(build_path):
            try:
                shutil.rmtree(build_path)
                mprint("<green>Successfully removed build folder:</green> {}", build_path)
            except Exception as e:
                mprint("<red>Error removing build folder:</red> {}", str(e))

    mprint("<green>Package '{}' has been removed.</green>", package_name)


"""
This file provides information of how to build and configure CLHEP library:
https://gitlab.cern.ch/CLHEP/CLHEP

"""

import os
import platform

from edpm.engine.generators.steps import EnvSet, EnvAppend, EnvPrepend
from edpm.engine.composed_recipe import ComposedRecipe


class ClhepRecipe(ComposedRecipe):
    """
    Installs the CLHEP library from GitLab.
    """
    def __init__(self, config):
        self.default_config = {
            'fetch': 'git',
            'make': 'cmake',
            'url': 'https://gitlab.cern.ch/CLHEP/CLHEP.git',
            'branch': 'CLHEP_2_4_7_1'
        }
        super().__init__(name='clhep', config=config)

    @staticmethod
    def gen_env(data):
        path = data['install_path']
        lib_path = os.path.join(path, 'lib')
        bin_path = os.path.join(path, 'bin')

        yield EnvSet('CLHEP', path)
        yield EnvSet('CLHEP_BASE_DIR', path)
        yield EnvSet('CLHEP_INCLUDE_DIR', os.path.join(path, 'include'))
        yield EnvSet('CLHEP_LIB_DIR', lib_path)

        yield EnvPrepend('PATH', bin_path)
        yield EnvPrepend('LD_LIBRARY_PATH', lib_path)

        if platform.system() == 'Darwin':
            yield EnvAppend('DYLD_LIBRARY_PATH', lib_path)




# tests/test_planfile.py
import pytest
import os
from ruamel.yaml import YAML
from tempfile import TemporaryDirectory

from edpm.engine.planfile import PlanFile, PlanPackage

def test_planfile_init_empty():
    pf = PlanFile({})
    assert "global" in pf.data
    assert isinstance(pf.data["packages"], list)
    assert pf.data["packages"] == []

    # default config
    cfg = pf.global_config()
    assert cfg["build_threads"] == 4
    assert cfg["cxx_standard"] == 17
    assert pf.data["global"]["environment"] == []

def test_planfile_load_and_save():
    with TemporaryDirectory() as tmpdir:
        plan_path = os.path.join(tmpdir, "plan.edpm.yaml")

        # Create a minimal plan file
        yaml = YAML()
        initial_data = {
            "global": {
                "config": {
                    "build_threads": 8,
                    "cxx_standard": 20
                },
                "environment": []
            },
            "packages": ["root"]
        }
        with open(plan_path, "w", encoding="utf-8") as f:
            yaml.dump(initial_data, f)

        # Load
        pf = PlanFile.load(plan_path)
        assert pf.data["global"]["config"]["build_threads"] == 8
        assert pf.data["packages"] == ["root"]

        # Modify
        pf.data["packages"].append("geant4@v11.03")
        pf.save(plan_path)

        # Reload
        pf2 = PlanFile.load(plan_path)
        assert pf2.data["packages"] == ["root", "geant4@v11.03"]

def test_planfile_packages():
    raw_data = {
        "global": {
            "config": {
                "build_threads": 8
            },
            "environment": []
        },
        "packages": [
            "root",
            "geant4@v11.03",
            {
                "mylib": {
                    "fetch": "git",
                    "branch": "main",
                    "environment": [
                        {"prepend": {"PATH": "$install_dir/bin"}},
                    ]
                }
            }
        ]
    }
    pf = PlanFile(raw_data)
    pkgs = pf.packages()
    assert len(pkgs) == 3

    # root
    p0 = pkgs[0]
    assert p0.name == "root"
    assert p0.is_baked_in
    assert p0.config == {}

    # geant4@v11.03
    p1 = pkgs[1]
    assert p1.name == "geant4"
    assert p1.is_baked_in
    assert p1.config == {"version": "v11.03"}

    # mylib
    p2 = pkgs[2]
    assert p2.name == "mylib"
    assert not p2.is_baked_in
    assert p2.config["fetch"] == "git"
    assert p2.config["branch"] == "main"
    env_steps = p2.env_block().parse()
    assert len(env_steps) == 1
    # We won't check the entire step type here, just ensure it parsed.

def test_planfile_add_package():
    pf = PlanFile({})
    assert pf.data["packages"] == []

    # Add a string
    pf.add_package("root")
    assert pf.data["packages"] == ["root"]

    # Add a dictionary
    pf.add_package({"geant4": {"fetch": "git"}})
    assert len(pf.data["packages"]) == 2
    assert pf.data["packages"][1] == {"geant4": {"fetch": "git"}}

    # Confirm we can parse them into PlanPackage objects
    pkgs = pf.packages()
    assert len(pkgs) == 2
    assert pkgs[0].name == "root"
    assert pkgs[1].name == "geant4"
    assert pkgs[1].config == {"fetch": "git"}

def test_planfile_has_and_find():
    raw_data = {
        "packages": ["root", "geant4@v11.03", {"mylib": {"fetch": "git"}}]
    }
    pf = PlanFile(raw_data)

    assert pf.has_package("root") is True
    assert pf.has_package("geant4") is True
    assert pf.has_package("mylib") is True
    assert pf.has_package("foo") is False

    p = pf.find_package("geant4")
    assert p is not None
    assert p.name == "geant4"
    assert p.config["version"] == "v11.03"



# tests/test_lockfile.py

import os
import pytest
import tempfile
from unittest.mock import patch
from edpm.engine.lockfile import LockfileConfig

def test_lockfile_init():
    """Test initialization of LockfileConfig."""
    lock = LockfileConfig()
    assert lock.file_path == ""
    assert lock.data["file_version"] == LockfileConfig.DEFAULT_FILE_VERSION
    assert lock.data["top_dir"] == ""
    assert lock.data["packages"] == {}
    assert lock.is_loaded == False

def test_lockfile_load_save():
    """Test loading and saving a lock file."""
    with tempfile.NamedTemporaryFile(suffix='.yaml', delete=False) as tmp:
        tmp_path = tmp.name

    try:
        # Create and save a lock file
        lock = LockfileConfig()
        lock.top_dir = "/test/dir"
        lock.data["packages"]["test"] = {"install_path": "/test/install"}
        lock.save(tmp_path)

        # Load it back and check
        lock2 = LockfileConfig()
        lock2.load(tmp_path)
        assert lock2.top_dir == "/test/dir"
        assert "test" in lock2.data["packages"]
        assert lock2.data["packages"]["test"]["install_path"] == "/test/install"
    finally:
        if os.path.exists(tmp_path):
            os.unlink(tmp_path)

def test_lockfile_get_installed_package():
    """Test getting an installed package."""
    lock = LockfileConfig()
    lock.data["packages"]["test"] = {"install_path": "/test/install"}

    package = lock.get_installed_package("test")
    assert package["install_path"] == "/test/install"

    # Non-existent package should return empty dict
    assert lock.get_installed_package("nonexistent") == {}

def test_lockfile_is_installed():
    """Test checking if a package is installed."""
    lock = LockfileConfig()

    # Mock os.path.isdir to always return True for testing
    with patch('os.path.isdir', return_value=True):
        # Not installed yet
        assert lock.is_installed("test") == False

        # Add it with an install_path
        lock.data["packages"]["test"] = {"install_path": "/test/install"}
        assert lock.is_installed("test") == True

        # Empty install_path should return False
        lock.data["packages"]["empty"] = {"install_path": ""}
        assert lock.is_installed("empty") == False

def test_lockfile_update_package():
    """Test updating a package."""
    lock = LockfileConfig()

    # Update a non-existent package
    lock.update_package("test", {"install_path": "/test/install"})
    assert "test" in lock.data["packages"]
    assert lock.data["packages"]["test"]["install_path"] == "/test/install"

    # Update an existing package
    lock.update_package("test", {"version": "1.0"})
    assert lock.data["packages"]["test"]["install_path"] == "/test/install"
    assert lock.data["packages"]["test"]["version"] == "1.0"

def test_lockfile_get_installed_packages():
    """Test getting all installed packages."""
    lock = LockfileConfig()
    assert lock.get_installed_packages() == []

    lock.data["packages"]["test1"] = {}
    lock.data["packages"]["test2"] = {}

    packages = lock.get_installed_packages()
    assert len(packages) == 2
    assert "test1" in packages
    assert "test2" in packages

def test_lockfile_remove_package():
    """Test removing a package."""
    lock = LockfileConfig()
    lock.data["packages"]["test"] = {"install_path": "/test/install"}

    assert "test" in lock.data["packages"]
    lock.remove_package("test")
    assert "test" not in lock.data["packages"]

    # Removing a non-existent package should not raise an error
    lock.remove_package("nonexistent")  # Should not raise any exception

def test_lockfile_save_no_path():
    """Test saving without a path raises an error."""
    lock = LockfileConfig()
    with pytest.raises(ValueError, match="No file path to save lockfile"):
        lock.save()

def test_lockfile_top_dir_property():
    """Test the top_dir property getter and setter."""
    lock = LockfileConfig()
    assert lock.top_dir == ""

    lock.top_dir = "/new/path"
    assert lock.top_dir == "/new/path"
    assert lock.data["top_dir"] == "/new/path"

def test_lockfile_load_nonexistent_file():
    """Test loading a non-existent file."""
    lock = LockfileConfig()
    # Use a temp path that definitely doesn't exist
    with tempfile.TemporaryDirectory() as tmpdir:
        nonexistent_path = os.path.join(tmpdir, "nonexistent.yaml")
        lock.load(nonexistent_path)
        assert lock.file_path == nonexistent_path
        assert lock.is_loaded == False


"""
This file provides information of how to build and configure HepMC framework:
https://gitlab.cern.ch/hepmc/HepMC3
"""
import os
import platform

from edpm.engine.generators.steps import EnvSet, EnvAppend, EnvPrepend, CmakePrefixPath
from edpm.engine.composed_recipe import ComposedRecipe


class HepMC3Recipe(ComposedRecipe):
    """
    Installs HepMC3 from Git + CMake.
    """
    def __init__(self, config):
        self.default_config = {
            'fetch': 'git',
            'make': 'cmake',
            'url': 'https://gitlab.cern.ch/hepmc/HepMC3.git',
            'branch': 'v3.3.0'
        }
        super().__init__(name='hepmc3', config=config)

    @staticmethod
    def gen_env(data):
        path = data['install_path']
        bin_path = os.path.join(path, 'bin')

        yield EnvPrepend('PATH', bin_path)
        yield EnvSet('HEPMC3_DIR', path)
        yield CmakePrefixPath(os.path.join(path, 'share', 'HepMC3', 'cmake'))

        if platform.system() == 'Darwin':
            yield EnvAppend('DYLD_LIBRARY_PATH', os.path.join(path, 'lib'))
        yield EnvAppend('LD_LIBRARY_PATH', os.path.join(path, 'lib'))



"""
Catch2 unit testing framework
https://github.com/catchorg/Catch2
"""

import os
import platform

from edpm.engine.generators.steps import EnvAppend, CmakePrefixPath
from edpm.engine.composed_recipe import ComposedRecipe


class Catch2Recipe(ComposedRecipe):
    """
    Installs the Catch2 C++ unit testing framework from GitHub.
    """
    def __init__(self, config):
        self.default_config = {
            'fetch': 'git',
            'make': 'cmake',
            # This can be adjusted if you prefer a stable branch like 'v3.3.2'
            'url': 'https://github.com/catchorg/Catch2.git',
            'branch': 'v3.8.1'
        }
        super().__init__(name='catch2', config=config)

    @staticmethod
    def gen_env(data):
        path = data['install_path']

        if platform.system() == 'Darwin':
            yield EnvAppend('DYLD_LIBRARY_PATH', os.path.join(path, 'lib'))

        yield EnvAppend('LD_LIBRARY_PATH', os.path.join(path, 'lib'))
        yield CmakePrefixPath(os.path.join(path, 'lib', 'cmake', 'Catch2'))



"""
Algorithms library:
https://github.com/eic/algorithms

"""
import os
import platform

from edpm.engine.generators.steps import EnvAppend, CmakePrefixPath
from edpm.engine.composed_recipe import ComposedRecipe


class AlgorithmsRecipe(ComposedRecipe):
    """
    Installs a hypothetical EIC 'algorithms' project from Git + CMake.
    """
    def __init__(self, config):
        self.default_config = {
            'fetch': 'git',
            'make': 'cmake',
            'url': 'https://github.com/eic/algorithms.git',
            'branch': 'v1.0.0'
        }
        super().__init__(name='algorithms', config=config)

    @staticmethod
    def gen_env(data):
        path = data['install_path']
        CmakePrefixPath(path)

        # On macOS, add to DYLD_LIBRARY_PATH
        if platform.system() == 'Darwin':
            yield EnvAppend('DYLD_LIBRARY_PATH', os.path.join(path, 'lib'))

        # On Linux or others, LD_LIBRARY_PATH
        yield EnvAppend('LD_LIBRARY_PATH', os.path.join(path, 'lib'))




"""
This file provides information of how to build and configure Fast Jet:
http://fastjet.fr

"""
import os

import platform
from edpm.engine.composed_recipe import ComposedRecipe
from edpm.engine.generators.steps import EnvSet, EnvPrepend, EnvAppend

class FastJetRecipe(ComposedRecipe):
    """
    Installs FastJet from a tarball using Autotools-like configure & make.
    Default is FastJet 3.4.3 from fastjet.fr.
    """

    def __init__(self, config):
        # Fill in default_config with the tarball URL and use 'autotools' build
        self.default_config = {
            'fetch': 'tarball',
            'make': 'autotools',
            'url': 'https://fastjet.fr/repo/fastjet-3.4.3.tar.gz',
            # If you need custom configure flags, e.g. debug or special options:
            # 'configure_flags': '--enable-allfeatures',
            # You can also set default 'build_threads': 8 if you want
        }
        super().__init__(name='fastjet', config=config)

    @staticmethod
    def gen_env(data):
        """
        Sets environment variables so that other software can discover FastJet.
        Mirrors the original environment logic from the older v2 recipe.
        """
        install_path = data['install_path']
        yield EnvPrepend('PATH', os.path.join(install_path, 'bin'))
        yield EnvPrepend('LD_LIBRARY_PATH', os.path.join(install_path, 'lib'))
        yield EnvSet('FASTJET_ROOT', install_path)

        if platform.system() == 'Darwin':
            yield EnvAppend('DYLD_LIBRARY_PATH', os.path.join(install_path, 'lib'))




"""
Nuclear Physics Detectors library
https://eicweb.phy.anl.gov/EIC/NPDet.git
"""
import os
import platform

from edpm.engine.generators.steps import EnvPrepend
from edpm.engine.composed_recipe import ComposedRecipe


class NpDetRecipe(ComposedRecipe):
    """
    Installs npdet from Git + CMake (for NP detectors).
    """
    def __init__(self, config):
        self.default_config = {
            'fetch': 'git',
            'make': 'cmake',
            'url': 'https://eicweb.phy.anl.gov/EIC/NPDet.git',
            'branch': 'v1.4.1'
        }
        super().__init__(name='npdet', config=config)

    @staticmethod
    def gen_env(data):
        path = data['install_path']

        yield EnvPrepend('LD_LIBRARY_PATH', os.path.join(path, 'lib'))
        yield EnvPrepend('PATH', os.path.join(path, 'bin'))

        if platform.system() == 'Darwin':
            yield EnvPrepend('DYLD_LIBRARY_PATH', os.path.join(path, 'lib'))

    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu': "space separated packet names", 'centos': "..."}
    # The idea behind is to generate easy to use instructions: 'sudo apt-get install ... ... ... '
    os_dependencies = {
        'required': {
            'ubuntu18': "libspdlog-dev libocct-foundation-dev occt-misc libocct-draw-dev libocct-data-exchange-dev libfmt-dev libtbb-dev",
            'ubuntu22': "libspdlog-dev libocct-foundation-dev occt-misc libocct-draw-dev libocct-data-exchange-dev libfmt-dev libtbb-dev",
        },
        'optional': {},
    }


import os
import click

from edpm.engine.api import EdpmApi
from edpm.engine.output import markup_print as mprint

@click.command("add", help="Add a new dependency entry to the plan file.")
@click.option("--fetch", default="", help="Fetcher type or URL (git/tarball/filesystem or autodetect from URL).")
@click.option("--make", default="", help="Maker type (cmake/autotools/manual/custom).")
@click.option("--branch", default="", help="Branch/tag (main, master, v1.2, etc.) if git fetcher.")
@click.option("--location", default="", help="Location/path if manual or filesystem fetcher.")
@click.option("--url", default="", help="Repo or tarball URL if fetch=git or fetch=tarball.")
@click.option("--existing", "-e", default="", help="Path to existing package installation.")
@click.option("--option", "option_list", multiple=True,
              help="Arbitrary key=value pairs that go into dependency config. E.g. --option cxx_standard=17.")
@click.argument("raw_name", required=True)
@click.pass_context
def add_command(ctx, raw_name, fetch, make, branch, location, url, existing, option_list):
    """
    Updates the plan.edpm.yaml "packages" list to include a new entry.

    Examples:
      edpm add root
      edpm add root@v6.32.0
      edpm add mylib --fetch=git --url=...
      edpm add --existing root /root/installation/path
    """
    api: EdpmApi = ctx.obj

    # Ensure plan is loaded or show an error
    if not api.plan:
        try:
            api.load_all()
        except FileNotFoundError:
            click.echo(
                "No plan file found. Please run 'edpm init' or specify --plan=... \n"
                "Cannot add dependency because the plan is not available."
            )
            raise click.Abort()
        if not api.plan:
            click.echo(
                "Error: EDPM plan data is not available (api.plan is still None). "
                "Please ensure plan is initialized."
            )
            raise click.Abort()

    # Split out the optional version part. e.g. "root@v6.32.0" => pkg_name="root", version_part="v6.32.0"
    pkg_name = raw_name
    version_part = ""
    if '@' in raw_name:
        parts = raw_name.split('@', 1)
        pkg_name = parts[0]
        version_part = parts[1]

    # Check if this package name already exists
    if api.plan.has_package(pkg_name):
        mprint("<red>Error:</red> A package named '{}' already exists in the plan.", pkg_name)
        raise click.Abort()

    # Check if it's a known recipe
    is_known_recipe = (pkg_name in api.recipe_manager.recipes_by_name)

    # Consider all flags, including the new 'existing' flag
    any_flags = any([fetch, make, branch, location, url, existing, option_list])

    # Decide how to store the item in 'packages'
    # 1) If known recipe, no other flags, and possibly version_part => single string
    if is_known_recipe and not any_flags:
        if version_part:
            # e.g. "root@v6.32.0"
            new_entry = f"{pkg_name}@{version_part}"
        else:
            # e.g. "root"
            new_entry = pkg_name
    else:
        # Otherwise produce a dictionary
        new_entry = {pkg_name: {}}
        config_block = new_entry[pkg_name]

        # If we have a version from '@', store it in 'version'
        if version_part:
            config_block["version"] = version_part

        # Handle --existing flag
        if existing:
            config_block["existing"] = existing

        # Populate fetch
        if fetch:
            config_block["fetch"] = fetch
        elif url:
            # autodetect fetch type from url if user didn't specify
            if url.endswith(".git"):
                config_block["fetch"] = "git"
            elif url.endswith(".tar.gz") or url.endswith(".tgz"):
                config_block["fetch"] = "tarball"
            else:
                config_block["fetch"] = "filesystem"

        # Populate URL or location
        if url:
            config_block["url"] = url
        if location:
            config_block["location"] = location

        # Make
        if make:
            config_block["make"] = make

        # Branch
        if branch:
            config_block["branch"] = branch

        # Additional options
        for opt in option_list:
            if "=" in opt:
                k, v = opt.split("=", 1)
                config_block[k.strip()] = v.strip()
            else:
                mprint("<yellow>Warning:</yellow> Ignoring malformed --option '{}'; expected key=value.", opt)

    # Append and save
    api.plan.data["packages"].append(new_entry)
    api.plan.save(api.plan_file)

    mprint("<green>Added dependency</green> '{}' to the plan.\nCheck plan-file to see or edit details:\n{}",
           raw_name, api.plan_file)


# edpm/engine/lockfile.py

import os
from typing import Dict, Any
from ruamel.yaml import YAML

yaml_rt = YAML(typ='rt')

class LockfileConfig:
    DEFAULT_FILE_VERSION = 1

    def __init__(self):
        self.file_path: str = ""
        self.data: Dict[str, Any] = {
            "file_version": self.DEFAULT_FILE_VERSION,
            "top_dir": "",
            "packages": {}
        }
        self.is_loaded = False

    def load(self, filepath: str):
        if not os.path.isfile(filepath):
            self.file_path = filepath
            return
        with open(filepath, "r", encoding="utf-8") as f:
            raw = yaml_rt.load(f) or {}
        self.data = raw
        self.file_path = filepath
        self.is_loaded = True

    def save(self, filepath: str = ""):
        if filepath:
            self.file_path = filepath
        if not self.file_path:
            raise ValueError("No file path to save lockfile.")
        with open(self.file_path, "w", encoding="utf-8") as f:
            yaml_rt.dump(self.data, f)

    @property
    def top_dir(self) -> str:
        return self.data.get("top_dir", "")

    @top_dir.setter
    def top_dir(self, path: str):
        self.data["top_dir"] = path

    def get_installed_package(self, name: str) -> Dict[str, Any]:
        return self.data["packages"].get(name, {})

    def is_installed(self, name: str) -> bool:
        dep_data = self.get_installed_package(name)
        ipath = dep_data.get("install_path", "")
        return bool(ipath and os.path.isdir(ipath))

    def update_package(self, name: str, info: Dict[str, Any]):
        if name not in self.data["packages"]:
            self.data["packages"][name] = {}
        self.data["packages"][name].update(info)

    def get_installed_packages(self):
        return list(self.data["packages"].keys())

    def remove_package(self, name: str):
        """
        Removes a package from the lock file.
        Note: This method silently ignores attempts to remove non-existent packages.
        """
        if name in self.data["packages"]:
            del self.data["packages"][name]


# edpm/engine/recipe.py

import os
from typing import Optional, List
from edpm.engine.config import ConfigNamespace

class Recipe:
    """
    Base class for all recipes.

    Recipes in EDPM define how to build, install, and manage a specific software package.
    This class provides a common structure and defines the lifecycle stages
    for all recipes, whether they are baked-in (simple recipes that inherit this class and override methods)
    or composed (selecting components like fetcher and maker in ComposedRecipe).

    """

    default_config = {}     # inherited classes can add values here before calling super().__init__(...).
    # This simplifies __init__() for inherited classes

    def __init__(self, config: Optional[ConfigNamespace]):

        # Smart copy of default configs into the config
        if self.default_config:
            local_config = self.default_config.copy()
            local_config.update(config)     # If user provided some fields, they should overwrite defaults
            config.update(local_config)     # This copies back whatever we have with config

        self.config = config if config else ConfigNamespace()
        # Next variables are set by ancestors


    def preconfigure(self):
        """This function is used to format and fill variables, when app_path is known download command
        This method is called before any other recipe steps (fetch, build, install, etc.).
        """
        pass

        def preconfigure(self):
            """
            Prepare configuration variables after app_path is known but before fetching.
            Used to format path templates and set up derived configuration values.
            """
        pass

    def fetch(self):
        """Download or clone the package sources to fetch_path"""
        pass

    def patch(self):
        """Apply any patches or source modifications to source_path"""
        pass

    def build(self):
        """Configure and compile the package in build_path"""
        pass

    def install(self):
        """Install the built package to install_path"""
        pass

    def post_install(self):
        """Perform post-installation tasks and verification"""
        pass

    def run_full_pipeline(self):
        """
        Execute the complete installation pipeline:
        1. fetch() -> 2. patch() -> 3. build() -> 4. install() -> 5. post_install()
        """
        self.fetch()
        self.patch()
        self.build()
        self.install()
        self.post_install()

    def use_common_dirs_scheme(self):
        """Function sets common directory scheme. It is the same for many packets:
        """
        if 'app_path' in self.config.keys():
            # where we download the source or clone git
            if "fetch_path" not in self.config.keys():
                self.config["fetch_path"] = "{app_path}/src".format(**self.config)

            # The directory with source files for current version
            if "source_path" not in self.config.keys():
                self.config["source_path"] = "{app_path}/src".format(**self.config)

            # The directory for cmake build
            if "build_path" not in self.config.keys():
                self.config["build_path"] = "{app_path}/build".format(**self.config)

            # The directory, where binary is installed
            if "install_path" not in self.config.keys():
                self.config["install_path"] = "{app_path}/{app_name}-install".format(**self.config)



import os
import click

from edpm.engine.api import EdpmApi, print_packets_info
from edpm.engine.output import markup_print as mprint
from edpm.version import version

# CLI Commands from your submodules
from edpm.cli.env import env_group
from edpm.cli.install import install_command
from edpm.cli.req import req_command
from edpm.cli.rm import rm_command
from edpm.cli.pwd import pwd_command
from edpm.cli.clean import clean_command
from edpm.cli.info import info_command
from edpm.cli.config import config_command
from edpm.cli.init import init_command
from edpm.cli.add import add_command

def print_first_time_message():
    mprint(
        """
No plan file found. It appears you're running 'edpm' for the first time.

Quick Start Guide:

1. Initialize a new plan file:
   > edpm init                   # Creates plan.edpm.yaml

2. Add and install packages:
   > edpm add clhep root         # Add a packages to the plan
   > edpm install                # Install the packages

3. By default, EDPM installs packages to './edpm-packages/'. 
   To use a different location:
   > edpm --top-dir=/path/to/storage
   This path is saved in the lock file for future use.

4. For pre-existing packages:
   > edpm add --existing root /path/to/existing/installation
   This registers external packages without reinstalling them.

Commands for packages:
   > edpm config <package>       # Set custom build flags
   > edpm pwd <package>          # Show installed package paths
   > edpm rm <package>           # Remove a package

For more help: edpm --help
Project repository: https://gitlab.com/DraTeots/edpm
"""
    )
    click.echo()


@click.group(invoke_without_command=True)
@click.option('--plan', default="", help="The plan file. Default is plan.edpm.yaml")
@click.option('--lock', default="", help="The lock file. Default is plan-lock.edpm.yaml")
@click.option('--top-dir', default="", help="Where EDPM should install missing packages.")
@click.pass_context
def edpm_cli(ctx, plan, lock, top_dir):
    """
    EDPM stands for EIC Development Packet Manager.
    If you run this command with no subcommand, it prints the version
    and a short summary of installed/known packages.
    """
    assert isinstance(ctx, click.Context), "EdpmApi context not available."

    # Get plan file path from environment variable or use default
    plan_file = os.environ.get("EDPM_PLAN_FILE", "plan.edpm.yaml")

    # Override with provided parameter if it exists
    if plan:
        plan_file = str(plan)

    # Determine lock file path
    if lock:
        # Explicit lock file path provided
        lock_file = str(lock)
    elif "EDPM_LOCK_FILE" in os.environ:
        # Environment variable for lock file
        lock_file = os.environ["EDPM_LOCK_FILE"]
    else:
        # Derive lock file path from plan file path
        plan_basename = os.path.basename(plan_file)
        plan_dir = os.path.dirname(plan_file)

        # Replace extension or add '-lock' before extension
        if '.' in plan_basename:
            name_part, ext = os.path.splitext(plan_basename)
            lock_basename = f"{name_part}-lock{ext}"
        else:
            lock_basename = f"{plan_basename}-lock.edpm.yaml"

        lock_file = os.path.join(plan_dir, lock_basename)

    if not os.path.isfile(plan_file) and ctx.invoked_subcommand != "init":
        print(f"Plan file does not exists (or there is no access to it): {plan_file}")
        click.echo("Running init command.")
        print_first_time_message()
        exit(1)

    api = EdpmApi(plan_file, lock_file)
    ctx.obj = api

    # Init command presumes there is no plan file. All other commands mean - we must load whatever we can
    if ctx.invoked_subcommand != "init":
        api.load_all()

    # Load db and modules from disk

    # If user passed --top-dir, set it in the DB
    if top_dir:
        if ctx.invoked_subcommand == "init":
            mprint("<b><red>ERROR</red></b> --top-dir flag is given with 'init' command, "
                   "which means the desired lock file doesn't exist yet and we can't save top-dir value. "
                   "Please run 'edpm init' without --top-dir flag and then use:\n"
                   f"edpm --top-dir={top_dir}\n\n")
            exit(1)
        api.lock.top_dir = os.path.abspath(os.path.normpath(top_dir))
        api.lock.save()

    # If no subcommand, print version and some package info
    if ctx.invoked_subcommand is None:
        mprint("<b><blue>edpm</blue></b> v{}", version)
        mprint("<b><blue>top dir :</blue></b>\n  {}", api.lock.top_dir)
        mprint("<b><blue>plan and lock:</blue></b>\n  {}\n  {}", api.plan_file, api.lock_file)
        print_packets_info(api)


# Register all subcommands
edpm_cli.add_command(install_command)
edpm_cli.add_command(env_group)
edpm_cli.add_command(req_command)
edpm_cli.add_command(rm_command)
edpm_cli.add_command(pwd_command)
edpm_cli.add_command(clean_command)
edpm_cli.add_command(info_command)
edpm_cli.add_command(config_command)
edpm_cli.add_command(init_command)
edpm_cli.add_command(add_command)



# edpm/engine/recipe_manager.py

import pkgutil
import importlib
from typing import Dict
from edpm.engine.recipe import Recipe
from edpm.engine.composed_recipe import ComposedRecipe   # We'll define it (see below)

def import_all_submodules(modules_dir, package_name):
    for (module_loader, name, ispkg) in pkgutil.iter_modules([modules_dir]):
        importlib.import_module('.' + name, package_name)

def all_subclasses(cls):
    results = set()
    for subclass in cls.__subclasses__():
        results.add(subclass)
        results.update(all_subclasses(subclass))
    return results

class RecipeManager:
    """
    Manages creation of 'baked in' recipes or 'composed' ones.
    """

    def __init__(self):
        self.recipes_by_name = {}
        # (Optional) keep track of known built-in names like "root", "geant4", etc.

    def load_installers(self, modules_dir=None, package_name="edpm.recipes"):
        """
        Import all recipe modules so that baked-in recipe classes are registered.
        For example, RootRecipe might set self.name = "root" internally.
        """
        if modules_dir is None:
            # Automatically determine the package directory
            package = importlib.import_module(package_name)
            modules_dir = package.__path__[0]

        import_all_submodules(modules_dir, package_name)

        # gather all classes that subclass Recipe
        classes = all_subclasses(Recipe)

        # go over classes
        for cls in classes:
            try:
                tmp_instance = cls(config={})  # e.g. RootRecipe()
                if hasattr(tmp_instance, 'name') and tmp_instance.name:
                    # store in a dictionary
                    rname = tmp_instance.name
                    self.recipes_by_name[rname] = cls
            except TypeError:
                pass

    def create_recipe(self, recipe_name, config: Dict[str, any]):
        """
        dep_obj is a planfile.PlanPackage instance.
        config is the merged config dict (global + local).
        Return a `Recipe` object (either baked-in or composed).
        """

        if recipe_name in self.recipes_by_name:
            recipe_cls = self.recipes_by_name[recipe_name]
            recipe = recipe_cls(config)  # instantiate
        else:
            # It's a custom / composed dependency
            # We'll create a ComposedRecipe
            recipe = ComposedRecipe(config=config, name=recipe_name)

        return recipe



"""
This file provides information of how to build and configure JANA2 packet:
https://github.com/JeffersonLab/JANA2

"""
import os
import platform

from edpm.engine.generators.steps import EnvSet, EnvAppend, EnvPrepend, CmakePrefixPath
from edpm.engine.composed_recipe import ComposedRecipe


class Jana2Recipe(ComposedRecipe):
    """
    Installs JANA2 from Git + CMake.
    """
    def __init__(self, config):
        self.default_config = {
            'fetch': 'git',
            'make': 'cmake',
            'url': 'https://github.com/JeffersonLab/JANA2.git',
            'branch': 'v2.4.0'
        }
        super().__init__(name='jana2', config=config)

    @staticmethod
    def gen_env(data):
        path = data['install_path']

        yield EnvSet('JANA_HOME', path)
        yield EnvAppend('JANA_PLUGIN_PATH', os.path.join(path, 'plugins'))
        yield EnvPrepend('PATH', os.path.join(path, 'bin'))
        yield EnvPrepend('LD_LIBRARY_PATH', os.path.join(path, 'lib'))
        yield EnvAppend('CMAKE_PREFIX_PATH', os.path.join(path, 'lib', 'cmake', 'JANA'))
        yield CmakePrefixPath(os.path.join(path, 'lib', 'cmake', 'JANA'))

        if platform.system() == 'Darwin':
            yield EnvAppend('DYLD_LIBRARY_PATH', os.path.join(path, 'lib'))



import os
import tempfile
import pytest
from unittest.mock import patch, MagicMock

from ruamel.yaml import YAML

from edpm.engine.api import EdpmApi
from edpm.engine.recipe_manager import RecipeManager
from edpm.cli.req import req_command, gather_requirements


class TestRecipe:
    name = "test_recipe"

    os_dependencies = {
        'required': {
            'ubuntu': "test-lib-dev test-common",
            'centos': "test-lib-devel test-common"
        },
        'optional': {
            'ubuntu': "test-optional",
            'centos': "test-optional"
        }
    }


def test_gather_requirements():
    """Test the gathering of requirements from recipes and plan file."""

    # Mock the EdpmApi and RecipeManager
    api = MagicMock()
    api.recipe_manager = MagicMock()
    api.plan = MagicMock()

    # Create a test package in the plan
    test_pkg = MagicMock()
    test_pkg.name = "test"
    test_pkg.config = {
        "require": {
            "ubuntu": {
                "required": ["plan-required-pkg"],
                "optional": ["plan-optional-pkg"],
            }
        }
    }

    # Mock finding the package
    api.plan.find_package = MagicMock(return_value=test_pkg)

    # Mock the recipe
    api.recipe_manager.recipes_by_name = {
        "test": TestRecipe
    }

    # Test gathering requirements
    required, optional = gather_requirements(api, "ubuntu", ("test",))

    # Verify results
    assert "plan-required-pkg" in required
    assert "test-lib-dev" in required
    assert "test-common" in required

    assert "plan-optional-pkg" in optional
    assert "test-optional" in optional

    # Required packages shouldn't be in optional
    assert "test-lib-dev" not in optional


def test_gather_requirements_global():
    """Test gathering requirements with global requirements."""

    # Mock the EdpmApi and RecipeManager
    api = MagicMock()
    api.recipe_manager = MagicMock()
    api.plan = MagicMock()

    # Mock global requirements
    api.plan.global_config = MagicMock(return_value={
        "require": {
            "ubuntu": {
                "required": ["global-required-pkg"],
                "optional": ["global-optional-pkg"]
            }
        }
    })

    # Create a test package in the plan
    test_pkg = MagicMock()
    test_pkg.name = "test"
    test_pkg.config = {}

    # Mock finding the package
    api.plan.find_package.return_value = test_pkg

    # Mock the recipe with no dependencies
    api.recipe_manager.recipes_by_name = {}

    # Test gathering requirements
    required, optional = gather_requirements(api, "ubuntu", ("test",))

    # Verify results
    assert "global-required-pkg" in required
    assert "global-optional-pkg" in optional


def test_gather_requirements_with_list():
    """Test gathering requirements when they're given as a list instead of dict."""

    # Mock the EdpmApi and RecipeManager
    api = MagicMock()
    api.recipe_manager = MagicMock()
    api.plan = MagicMock()

    # Create a test package in the plan
    test_pkg = MagicMock()
    test_pkg.name = "test"
    test_pkg.config = {
        "require": {
            "ubuntu": ["list-pkg1", "list-pkg2"]
        }
    }

    # Mock finding the package
    api.plan.find_package = MagicMock(return_value=test_pkg)

    # Mock the recipe
    api.recipe_manager.recipes_by_name = {}

    # Test gathering requirements
    required, optional = gather_requirements(api, "ubuntu", ("test",))

    # Verify results
    assert "list-pkg1" in required
    assert "list-pkg2" in required
    assert not optional  # Should be empty


import os
import pytest
import shutil
from unittest.mock import patch, MagicMock
from click.testing import CliRunner

from edpm.cli.clean import clean_command
from edpm.engine.api import EdpmApi
from edpm.engine.lockfile import LockfileConfig
import click

@click.group()
@click.pass_context
def cli(ctx):
    pass

cli.add_command(clean_command)

@pytest.fixture
def runner():
    return CliRunner()

import os
import pytest
from edpm.engine.api import EdpmApi
from edpm.engine.planfile import PlanFile
from edpm.engine.lockfile import LockfileConfig
from ruamel.yaml import YAML

@pytest.fixture
def mock_api(tmp_path):
    """
    Creates minimal plan + lock files on disk so 'api.load_all()' doesn't fail.
    Also sets up an 'installed' package named 'mypkg'.
    """
    plan_path = tmp_path / "plan.edpm.yaml"
    lock_path = tmp_path / "plan-lock.edpm.yaml"

    # Write a minimal plan file
    minimal_plan = {
        "global": {
            "config": {},
            "environment": []
        },
        "packages": []
    }
    yaml = YAML()
    with plan_path.open("w", encoding="utf-8") as f:
        yaml.dump(minimal_plan, f)

    # Create an EdpmApi that references those files
    api = EdpmApi(plan_file=str(plan_path), lock_file=str(lock_path))

    # Mock a lock file in memory
    lock = LockfileConfig()
    lock.file_path = str(lock_path)
    lock.data = {
        "top_dir": str(tmp_path),
        "packages": {
            "mypkg": {
                "install_path": str(tmp_path / "mypkg_install"),
                "built_with_config": {
                    "build_path": str(tmp_path / "mypkg_build"),
                    "source_path": str(tmp_path / "mypkg_src")
                },
                "owned": True
            }
        }
    }
    api.lock = lock

    # Create physical directories so isdir() sees them
    os.makedirs(tmp_path / "mypkg_install", exist_ok=True)
    os.makedirs(tmp_path / "mypkg_build", exist_ok=True)
    os.makedirs(tmp_path / "mypkg_src", exist_ok=True)

    return api

def test_clean_success(runner, mock_api):
    """
    Test that clean_command removes 'mypkg' directories from disk when owned = True.
    We mock shutil.rmtree to confirm it's called with the correct directories.
    """
    # 1) Capture directories from the lock before calling 'clean'
    mypkg_data = mock_api.lock.data["packages"]["mypkg"]
    install_dir = mypkg_data["install_path"]
    build_dir   = mypkg_data["built_with_config"]["build_path"]
    source_dir  = mypkg_data["built_with_config"]["source_path"]

    with patch.object(shutil, 'rmtree', side_effect=MagicMock()) as mock_rmtree, \
            patch.object(mock_api.lock, 'save', side_effect=MagicMock()) as mock_lock_save, \
            patch.object(mock_api, 'save_generator_scripts', side_effect=MagicMock()) as mock_save_gens:

        # 2) Invoke the command
        result = runner.invoke(cli, ["clean", "mypkg"], obj=mock_api)
        assert result.exit_code == 0
        assert "Cleaned 'mypkg'" in result.output

        # 3) Verify the calls
        # We expect rmtree calls for install_dir, build_dir, source_dir
        calls = [call.args[0] for call in mock_rmtree.call_args_list]
        assert build_dir in calls
        assert source_dir in calls

        mock_save_gens.assert_called_once()

def test_clean_missing_install(runner, mock_api):
    """
    If the install_path is empty or the folder is missing, we show an error but do not rmtree anything.
    """
    # remove the install directory
    install_dir = mock_api.lock.data["packages"]["mypkg"]["install_path"]
    shutil.rmtree(install_dir)  # physically remove it

    with patch.object(shutil, 'rmtree', side_effect=MagicMock()) as mock_rmtree:
        result = runner.invoke(cli, ["clean", "mypkg"], obj=mock_api)
        assert result.exit_code == 0
        assert "not currently installed" in result.output
        mock_rmtree.assert_not_called()

def test_clean_not_in_lock(runner, mock_api):
    """
    If user tries to clean a package that isn't in the lock file.
    """
    result = runner.invoke(cli, ["clean", "someOtherPkg"], obj=mock_api)
    assert result.exit_code != 0
    assert "No installation info found for 'someOtherPkg'" in result.output

def test_clean_not_owned(runner, mock_api):
    """
    If is_owned=False, we skip removing from disk, just warn user.
    """
    mock_api.lock.data["packages"]["mypkg"]["owned"] = False
    with patch.object(shutil, 'rmtree', side_effect=MagicMock()) as mock_rmtree:
        result = runner.invoke(cli, ["clean", "mypkg"], obj=mock_api)
        assert result.exit_code == 0
        assert "is not owned by EDPM. Remove manually" in result.output
        mock_rmtree.assert_not_called()



# tests/test_cli_cmd_config.py
import pytest
from click.testing import CliRunner
import click
from ruamel.yaml import YAML
import os

# Import your actual config command
from edpm.cli.config import config_command
# And the EdpmApi
from edpm.engine.api import EdpmApi

@click.group()
@click.pass_context
def cli(ctx):
    pass

# Add the config command
cli.add_command(config_command, "config")

@pytest.fixture
def runner():
    return CliRunner()

@pytest.fixture
def temp_plan(tmp_path):
    """
    Creates a minimal plan.edpm.yaml with an empty 'packages' list
    and an empty global config. Then instantiates an EdpmApi
    pointing to it, calling load_all() so 'api.plan' is not None.
    """
    plan_path = tmp_path / "plan.edpm.yaml"
    lock_path = tmp_path / "plan-lock.edpm.yaml"

    # Minimal plan
    initial_plan = {
        "global": {
            "config": {}
        },
        "packages": []
    }
    yaml = YAML()
    with plan_path.open("w", encoding="utf-8") as f:
        yaml.dump(initial_plan, f)

    # Create the EdpmApi
    api = EdpmApi(plan_file=str(plan_path), lock_file=str(lock_path))
    api.load_all()  # Ensures api.plan is now a valid PlanFile object

    return api, plan_path, lock_path

@pytest.mark.usefixtures("temp_plan")
def test_show_global_config(runner, temp_plan):
    api, plan_path, lock_path = temp_plan
    # No arguments => show global config
    result = runner.invoke(cli, ["config"], obj=api)
    assert result.exit_code == 0
    assert "Global Config" in result.output
    assert "No global config found" in result.output

@pytest.mark.usefixtures("temp_plan")
def test_set_global_config(runner, temp_plan):
    api, plan_path, _ = temp_plan
    result = runner.invoke(cli, ["config", "build_threads=4", "cxx_standard=17"], obj=api)
    assert result.exit_code == 0
    assert "Updated global config" in result.output
    assert "build_threads = 4" in result.output
    assert "cxx_standard = 17" in result.output
    assert "Saved changes to" in result.output

    # Reload from disk
    api.load_all()
    glob_cfg = api.plan.data["global"]["config"]
    assert glob_cfg["build_threads"] == "4"
    assert glob_cfg["cxx_standard"] == "17"

@pytest.mark.usefixtures("temp_plan")
def test_show_dep_not_found(runner, temp_plan):
    api, _, _ = temp_plan
    result = runner.invoke(cli, ["config", "fakeDep"], obj=api)
    assert result.exit_code == 0
    assert "No dependency named 'fakeDep' in the plan" in result.output

@pytest.mark.usefixtures("temp_plan")
def test_create_new_dep(runner, temp_plan):
    api, plan_path, _ = temp_plan
    # Provide 'recipe=git'
    result = runner.invoke(cli, ["config", "mydep", "recipe=git", "url=https://some.git"], obj=api)
    assert result.exit_code == 0
    assert "Updated config for mydep" in result.output
    assert "recipe = git" in result.output
    assert "url = https://some.git" in result.output

@pytest.mark.usefixtures("temp_plan")
def test_create_new_dep_no_recipe(runner, temp_plan):
    """
    The code should produce an error if user tries to create a
    brand-new dependency but does NOT specify recipe=...
    """
    api, _, _ = temp_plan
    result = runner.invoke(cli, ["config", "unicorn", "branch=magic"], obj=api)
    assert result.exit_code == 0
    # Should show the error (per the test expectation).
    assert "No dependency named 'unicorn' in the plan, and no 'recipe=...' provided" in result.output

@pytest.mark.usefixtures("temp_plan")
def test_convert_string_dep_to_dict(runner, temp_plan):
    """
    If packages: ["root"] and we do 'edpm config root branch=main',
    we convert "root" => { "root": { recipe: "root", branch: "main" } }
    """
    api, plan_path, _ = temp_plan

    # Manually store packages: ["root"]
    api.plan.data["packages"] = ["root"]
    api.plan.save(api.plan_file)

    result = runner.invoke(cli, ["config", "root", "branch=main"], obj=api)
    assert result.exit_code == 0
    assert "convert 'root' to dict" not in result.output  # optional
    assert "Updated config for root" in result.output
    # The code should mention "recipe = root" as well
    assert "recipe = root" in result.output
    assert "branch = main" in result.output

@pytest.mark.usefixtures("temp_plan")
def test_show_dep_dict(runner, temp_plan):
    api, plan_path, _ = temp_plan
    # Insert a dictionary-based dep
    api.plan.data["packages"] = [
        {"foo": {"recipe": "git", "branch": "dev"}}
    ]
    api.plan.save(api.plan_file)

    result = runner.invoke(cli, ["config", "foo"], obj=api)
    assert result.exit_code == 0
    assert "foo config:" in result.output
    assert "recipe: git" in result.output
    assert "branch: dev" in result.output

@pytest.mark.usefixtures("temp_plan")
def test_update_existing_dep(runner, temp_plan):
    api, plan_path, _ = temp_plan
    # Already have {"mylib": {"recipe": "git", "url": "old"}}
    api.plan.data["packages"] = [
        {"mylib": {"recipe": "git", "url": "https://old.git"}}
    ]
    api.plan.save(api.plan_file)

    result = runner.invoke(cli, ["config", "mylib", "url=https://new.git", "branch=featureX"], obj=api)
    assert result.exit_code == 0
    assert "Updated config for mylib" in result.output
    assert "url = https://new.git" in result.output
    assert "branch = featureX" in result.output

    # Confirm it's saved
    api.load_all()
    pkgs = api.plan.data["packages"]
    dep_dict = pkgs[0]["mylib"]
    assert dep_dict["recipe"] == "git"
    assert dep_dict["url"] == "https://new.git"
    assert dep_dict["branch"] == "featureX"



"""
This file provides information of how to build and configure VMC framework:
Geometry conversion tool, providing conversion between Geant4 and ROOT TGeo geometry models.

https://github.com/vmc-project/vgm


"""
import os
import platform

from edpm.engine.generators.steps import EnvSet, EnvAppend
from edpm.engine.composed_recipe import ComposedRecipe


class VgmRecipe(ComposedRecipe):
    """
    Installs VGM (a geometry conversion tool for Geant4/ROOT).
    """
    def __init__(self, config):
        self.default_config = {
            'fetch': 'git',
            'make': 'cmake',
            'url': 'https://github.com/vmc-project/vgm.git',
            'branch': 'v5-3-1'
        }
        super().__init__(name='vgm', config=config)

    def gen_env(self, data):
        path = data['install_path']

        yield EnvSet('VGM_DIR', path)

        if platform.system() == 'Darwin':
            yield EnvAppend('DYLD_LIBRARY_PATH', os.path.join(path, 'lib'))
            yield EnvAppend('DYLD_LIBRARY_PATH', os.path.join(path, 'lib64'))

        yield EnvAppend('LD_LIBRARY_PATH', os.path.join(path, 'lib'))
        yield EnvAppend('LD_LIBRARY_PATH', os.path.join(path, 'lib64'))



import click
from edpm.engine.output import markup_print as mprint


@click.command("install")
@click.option('--force', is_flag=True, default=False, help="Force rebuild/reinstall even if already installed.")
@click.option('--top-dir', default="", help="Override or set top_dir in the lock file.")
@click.option('--explain', 'just_explain', is_flag=True, default=False, help="Print what would be installed but don't actually install.")
@click.option('--add', '-a', is_flag=True, default=False, help="Automatically add packages to the plan if not already present.")
@click.argument('names', nargs=-1)
@click.pass_context
def install_command(ctx, names, add, top_dir, just_explain, force):
    """
    Installs packages (and their dependencies) from the plan, updating the lock file.

    Use Cases:
      1) 'edpm install' with no arguments installs EVERYTHING in the plan.
      2) 'edpm install <pkg>' adds <pkg> to the plan if not present, then installs it.
    """

    edpm_api = ctx.obj
    # assert isinstance(edpm_api, EdpmApi)

    # 2) Possibly override top_dir
    if top_dir:
        edpm_api.top_dir = top_dir

    # 3) If no arguments => install everything from the plan
    if not names:
        # "dep_names" = all from the plan
        dep_names = [dep.name for dep in edpm_api.plan.packages()]
        if not dep_names:
            mprint("<red>No dependencies in the plan!</red> "
                   "Please add packages or run 'edpm install <pkg>' to auto-add.")
            return
    else:
        # If user provided package names, let's auto-add them to the plan if not present
        # Parse package names to extract base names (without @version)
        dep_names = []
        for pkg_name in names:
            # Parse package name to extract base name and version
            base_name = pkg_name
            user_version = None
            if '@' in pkg_name:
                base_name, user_version = pkg_name.split('@', 1)

            dep_names.append(base_name)

            # Check if package is in plan (using base name)
            if not edpm_api.plan.has_package(base_name):
                if add:
                    # Auto-add the package to the plan with --add/-a flag
                    mprint(f"<yellow>Package '{pkg_name}' not in plan.</yellow> "
                           f"Adding it automatically (-a,--add flag)")
                    # Call the add_command logic to add the package
                    try:
                        # Add the full package name (with version) to the plan
                        edpm_api.plan.add_package(pkg_name)
                        edpm_api.plan.save(edpm_api.plan_file)
                        mprint(f"<green>Added '{pkg_name}' to the plan.</green>")
                    except Exception as e:
                        mprint(f"<red>Error:</red> Failed to add '{pkg_name}' to plan: {str(e)}")
                        exit(1)
                else:
                    # Without --add flag, show an error and suggest using it
                    _print_error_not_in_plan(pkg_name)
                    exit(1)
            else:
                # Package exists in plan - check for version conflicts
                if user_version:
                    existing_pkg = edpm_api.plan.find_package(base_name)
                    existing_version = existing_pkg.config.get("version", "")

                    if existing_version and existing_version != user_version:
                        _print_error_version_conflict(base_name, existing_version, user_version)
                        exit(1)

    # 4) Actually run the install logic
    edpm_api.install_dependency_chain(
        dep_names=dep_names,
        explain=just_explain,
        force=force
    )

    # 5) If not just_explain, optionally generate environment scripts
    if not just_explain:
        mprint("\nUpdating environment script files...\n")
        edpm_api.save_generator_scripts()


def _print_error_not_in_plan(pkg_name):
    mprint(f"<red>Error:</red> '{pkg_name}' is not in plan!")
    mprint(f"Options:")
    mprint(f"1. Add it to plan by editing the file")
    mprint(f"2. Use <blue>'edpm add {pkg_name}'</blue> command")
    mprint(f"3. Use <blue>'edpm install --add {pkg_name}'</blue> to add and install")


def _print_error_version_conflict(base_name, existing_version, user_version):
    mprint(f"<red>Error:</red> Version conflict for package '{base_name}'!")
    mprint(f"  Plan has version: <blue>{existing_version}</blue>")
    mprint(f"  You specified: <blue>{user_version}</blue>")
    mprint(f"")
    mprint(f"Options:")
    mprint(f"1. Edit the plan file to change the version")
    mprint(f"2. Use <blue>'edpm config {base_name} version={user_version}'</blue> to update the plan")
    mprint(f"3. Install without specifying version: <blue>'edpm install {base_name}'</blue>")



"""
Evio v6 JLab input output library
https://github.com/JeffersonLab/evio
"""

import os
import platform

from edpm.engine.generators.steps import EnvAppend, CmakePrefixPath, EnvSet, CmakeSet
from edpm.engine.composed_recipe import ComposedRecipe


class EvioRecipe(ComposedRecipe):
    """
    Installs the evio library
    """
    def __init__(self, config):
        self.default_config = {
            'fetch': 'git',
            'make': 'cmake',
            'url': 'https://github.com/JeffersonLab/evio',
            'branch': 'v6.0.0'
        }
        super().__init__(name='evio', config=config)

    @staticmethod
    def gen_env(data):
        path = data['install_path']

        if platform.system() == 'Darwin':
            yield EnvAppend('DYLD_LIBRARY_PATH', os.path.join(path, 'lib'))

        yield EnvAppend('LD_LIBRARY_PATH', os.path.join(path, 'lib'))
        yield CmakePrefixPath(os.path.join(path, 'lib', 'cmake'))
        yield EnvSet('DISRUPTOR_CPP_HOME', path)
        yield CmakeSet('DISRUPTOR_CPP_HOME', path)
        yield CmakeSet('DISRUPTOR_CPP_DIR', path)

    require = {
        "apt": ["liblz4-dev"]

    }


"""
JLab version of disruptor-cpp library
https://github.com/JeffersonLab/Disruptor-cpp
"""

import os
import platform

from edpm.engine.generators.steps import EnvAppend, CmakePrefixPath
from edpm.engine.composed_recipe import ComposedRecipe


class EvioRecipe(ComposedRecipe):
    """
    Installs the JLab version of disruptor-cpp library
    """
    def __init__(self, config):
        self.default_config = {
            'fetch': 'git',
            'make': 'cmake',
            'url': 'https://github.com/JeffersonLab/Disruptor-cpp.git',
            'branch': 'master'
        }
        super().__init__(name='disruptor-cpp', config=config)

    def preconfigure(self):
        # We need to do it in preconfigure when install_path is known
        self.config["cmake_flags"]=f"-DINSTALL_DIR={self.config['install_path']}"
        super().preconfigure()

    @staticmethod
    def gen_env(data):
        path = data['install_path']

        if platform.system() == 'Darwin':
            yield EnvAppend('DYLD_LIBRARY_PATH', os.path.join(path, 'lib'))

        yield EnvAppend('LD_LIBRARY_PATH', os.path.join(path, 'lib'))
        yield CmakePrefixPath(os.path.join(path, 'lib', 'cmake'))

    require = {
        "apt": ["liblz4-dev"]

    }


import os
import pytest
import click
from unittest.mock import MagicMock, patch
from click.testing import CliRunner
from edpm.cli.install import install_command
from edpm.engine.api import EdpmApi
from edpm.engine.planfile import PlanFile


@pytest.fixture
def mock_edpm_api():
    """Create a mock EdpmApi object with necessary attributes."""
    api = MagicMock(spec=EdpmApi)
    api.plan = MagicMock()
    api.plan_file = "test_plan.edpm.yaml"
    return api


@pytest.fixture
def cli_runner():
    """Return a Click CLI test runner."""
    return CliRunner()


def test_install_no_arguments(mock_edpm_api, cli_runner):
    """Test 'edpm install' with no arguments - should install all packages in plan."""
    # Setup mock API
    # Create proper mock packages with name attributes that return strings
    pkg1 = MagicMock()
    pkg1.name = "pkg1"
    pkg2 = MagicMock()
    pkg2.name = "pkg2"
    mock_edpm_api.plan.packages.return_value = [pkg1, pkg2]

    # We need to make the instance check pass by patching isinstance
    with patch('edpm.cli.install.isinstance', return_value=True):
        # Run the command
        result = cli_runner.invoke(install_command, obj=mock_edpm_api)

    # Verify
    assert result.exit_code == 0, f"Command failed with error: {result.output}"

    # Should call install_dependency_chain with all package names as a list
    mock_edpm_api.install_dependency_chain.assert_called_once_with(
        dep_names=["pkg1", "pkg2"],
        explain=False,
        force=False
    )
    # Should call save_generator_scripts
    mock_edpm_api.save_generator_scripts.assert_called_once()


def test_install_no_packages_in_plan(mock_edpm_api, cli_runner):
    """Test 'edpm install' when there are no packages in the plan."""
    # Setup mock API
    mock_edpm_api.plan.packages.return_value = []

    # We need to make the instance check pass by patching isinstance
    with patch('edpm.cli.install.isinstance', return_value=True):
        # Run the command
        result = cli_runner.invoke(install_command, obj=mock_edpm_api)

    # Verify
    assert result.exit_code == 0, f"Command failed with error: {result.output}"
    assert "No dependencies in the plan" in result.output

    # Should not call install_dependency_chain
    mock_edpm_api.install_dependency_chain.assert_not_called()


def test_install_with_specific_package(mock_edpm_api, cli_runner):
    """Test 'edpm install pkg1' with a package that exists in the plan."""
    # Setup mock API
    mock_edpm_api.plan.has_package.return_value = True

    # We need to make the instance check pass by patching isinstance
    with patch('edpm.cli.install.isinstance', return_value=True):
        # Run the command
        result = cli_runner.invoke(install_command, ["pkg1"], obj=mock_edpm_api)

    # Verify
    assert result.exit_code == 0, f"Command failed with error: {result.output}"

    # Should check if package exists
    mock_edpm_api.plan.has_package.assert_called_once_with("pkg1")

    # Should call install_dependency_chain with the specified package as a list
    mock_edpm_api.install_dependency_chain.assert_called_once_with(
        dep_names=["pkg1"],
        explain=False,
        force=False
    )
    # Should call save_generator_scripts
    mock_edpm_api.save_generator_scripts.assert_called_once()


def test_install_package_not_in_plan_without_add_flag(mock_edpm_api, cli_runner):
    """Test 'edpm install pkg1' with a package not in the plan without --add flag."""
    # Setup mock API
    mock_edpm_api.plan.has_package.return_value = False

    # Make the instance check pass
    with patch('edpm.cli.install.isinstance', return_value=True):
        # Run the command with capturing SystemExit
        result = cli_runner.invoke(install_command, ["pkg1"], obj=mock_edpm_api)

    # Verify (exit code 1 is OK here since the code calls exit(1))
    assert result.exit_code == 1, "Should exit with error code when package not in plan"
    assert "not in plan" in result.output
    assert "Options:" in result.output

    # Should not call install_dependency_chain
    mock_edpm_api.install_dependency_chain.assert_not_called()


def test_install_package_not_in_plan_with_add_flag(mock_edpm_api, cli_runner):
    """Test 'edpm install --add pkg1' with a package not in the plan."""
    # Setup mock API
    mock_edpm_api.plan.has_package.return_value = False

    # Make the instance check pass
    with patch('edpm.cli.install.isinstance', return_value=True):
        # Run the command
        result = cli_runner.invoke(install_command, ["--add", "pkg1"], obj=mock_edpm_api)

    # Verify
    assert result.exit_code == 0, f"Command failed with error: {result.output}"
    assert "Adding it automatically" in result.output

    # Should add package to plan
    mock_edpm_api.plan.add_package.assert_called_once_with("pkg1")
    mock_edpm_api.plan.save.assert_called_once_with(mock_edpm_api.plan_file)

    # Should call install_dependency_chain with a list
    mock_edpm_api.install_dependency_chain.assert_called_once_with(
        dep_names=["pkg1"],
        explain=False,
        force=False
    )


def test_install_multiple_packages_with_add_flag(mock_edpm_api, cli_runner):
    """Test 'edpm install --add pkg1 pkg2' with some packages not in the plan."""
    # Setup mock API
    # pkg1 exists, pkg2 doesn't
    mock_edpm_api.plan.has_package.side_effect = lambda pkg: pkg == "pkg1"

    # Make the instance check pass
    with patch('edpm.cli.install.isinstance', return_value=True):
        # Run the command
        result = cli_runner.invoke(install_command, ["--add", "pkg1", "pkg2"], obj=mock_edpm_api)

    # Verify
    assert result.exit_code == 0, f"Command failed with error: {result.output}"
    assert "Adding it automatically" in result.output

    # Should check both packages
    assert mock_edpm_api.plan.has_package.call_count == 2

    # Should add only pkg2 to plan
    mock_edpm_api.plan.add_package.assert_called_once_with("pkg2")
    mock_edpm_api.plan.save.assert_called_once_with(mock_edpm_api.plan_file)

    # Should call install_dependency_chain with both packages as a list
    mock_edpm_api.install_dependency_chain.assert_called_once_with(
        dep_names=["pkg1", "pkg2"],
        explain=False,
        force=False
    )


def test_install_with_explain_flag(mock_edpm_api, cli_runner):
    """Test 'edpm install --explain pkg1' which should not perform actual install."""
    # Setup mock API
    mock_edpm_api.plan.has_package.return_value = True

    # Make the instance check pass
    with patch('edpm.cli.install.isinstance', return_value=True):
        # Run the command
        result = cli_runner.invoke(install_command, ["--explain", "pkg1"], obj=mock_edpm_api)

    # Verify
    assert result.exit_code == 0, f"Command failed with error: {result.output}"

    # Should call install_dependency_chain with explain=True and a list
    mock_edpm_api.install_dependency_chain.assert_called_once_with(
        dep_names=["pkg1"],
        explain=True,
        force=False
    )

    # Should NOT call save_generator_scripts
    mock_edpm_api.save_generator_scripts.assert_not_called()


def test_install_with_force_flag(mock_edpm_api, cli_runner):
    """Test 'edpm install --force pkg1' which should force reinstall."""
    # Setup mock API
    mock_edpm_api.plan.has_package.return_value = True

    # Make the instance check pass
    with patch('edpm.cli.install.isinstance', return_value=True):
        # Run the command
        result = cli_runner.invoke(install_command, ["--force", "pkg1"], obj=mock_edpm_api)

    # Verify
    assert result.exit_code == 0, f"Command failed with error: {result.output}"

    # Should call install_dependency_chain with force=True and a list
    mock_edpm_api.install_dependency_chain.assert_called_once_with(
        dep_names=["pkg1"],
        explain=False,
        force=True
    )


def test_install_with_top_dir_flag(mock_edpm_api, cli_runner):
    """Test 'edpm install --top-dir=/custom/path pkg1' which should set custom top directory."""
    # Setup mock API
    mock_edpm_api.plan.has_package.return_value = True

    # Make the instance check pass
    with patch('edpm.cli.install.isinstance', return_value=True):
        # Run the command
        result = cli_runner.invoke(install_command, ["--top-dir", "/custom/path", "pkg1"], obj=mock_edpm_api)

    # Verify
    assert result.exit_code == 0, f"Command failed with error: {result.output}"

    # Should set top_dir
    assert mock_edpm_api.top_dir == "/custom/path"

    # Then proceed with install with a list
    mock_edpm_api.install_dependency_chain.assert_called_once_with(
        dep_names=["pkg1"],
        explain=False,
        force=False
    )


def test_install_add_package_failure(mock_edpm_api, cli_runner):
    """Test when adding a package with --add flag fails."""
    # Setup mock API
    mock_edpm_api.plan.has_package.return_value = False
    mock_edpm_api.plan.add_package.side_effect = Exception("Failed to add package")

    # Make the instance check pass
    with patch('edpm.cli.install.isinstance', return_value=True):
        # Run the command
        result = cli_runner.invoke(install_command, ["--add", "pkg1"], obj=mock_edpm_api)

    # Verify (should fail with exit code 1)
    assert result.exit_code == 1, "Command should exit with error code on package add failure"
    assert "Failed to add" in result.output

    # Should try to add package
    mock_edpm_api.plan.add_package.assert_called_once_with("pkg1")

    # Should not proceed with install
    mock_edpm_api.install_dependency_chain.assert_not_called()


"""
This file provides information of how to build and configure Eigen3 packet:
https://gitlab.com/libeigen/eigen.git
"""

import os

from edpm.engine.generators.steps import CmakePrefixPath
from edpm.engine.composed_recipe import ComposedRecipe


class EigenRecipe(ComposedRecipe):
    """Provides data for building and installing Eicgen3 framework"""

    def __init__(self, config):

        # Default values for the recipe
        self.default_config = {
            'fetch': 'git',
            'make': 'cmake',
            'branch': '3.4.0',
            'url': 'https://gitlab.com/libeigen/eigen.git'
        }
        super().__init__(name='eigen3', config=config)


    @staticmethod
    def gen_env(data):
        """Generates environments to be set"""
        yield CmakePrefixPath(os.path.join(data['install_path'], 'share/eigen3/cmake/'))




"""
Indirect Ray Tracing code for EPIC event reconstruction
https://github.com/eic/irt.git

"""
import os
import platform

from edpm.engine.generators.steps import EnvPrepend, CmakePrefixPath
from edpm.engine.composed_recipe import ComposedRecipe


class IrtRecipe(ComposedRecipe):
    """
    Installs IRT (Imaging Reconstruction Toolkit) from Git + CMake.
    """
    def __init__(self, config):
        self.default_config = {
            'fetch': 'git',
            'make': 'cmake',
            'url': 'https://github.com/eic/irt.git',
            'branch': 'v1.0.8'
        }
        super().__init__(name='irt', config=config)

    @staticmethod
    def gen_env(data):
        path = data['install_path']

        yield CmakePrefixPath(os.path.join(path, 'lib', 'cmake', 'IRT'))
        yield EnvPrepend('LD_LIBRARY_PATH', os.path.join(path, 'lib'))
        if platform.system() == 'Darwin':
            yield EnvPrepend('DYLD_LIBRARY_PATH', os.path.join(path, 'lib'))




import os
from abc import ABC, abstractmethod
from typing import Dict, Any
from edpm.engine.commands import run, workdir


# -------------------------------------
# F E T C H E R   I N T E R F A C E
# -------------------------------------

class IFetcher(ABC):
    """ Interface for "fetchers" that handle retrieving source code, etc. """

    def __init__(self, config: Dict[str, Any]):
        self.config = config

    def preconfigure(self):
        """
        This function can be used to add or refine config entries
        needed before the actual fetch step. Default: do nothing.
        """
        pass

    @abstractmethod
    def fetch(self):
        """
        Actually perform the fetch/cloning/copying step
        (e.g. git clone, tarball download+extract, etc.).
        """
        pass


class GitFetcher(IFetcher):
    """
    A fetcher that uses `git clone`.
    """

    def preconfigure(self):
        """
        EnvSet up config['clone_command'] if not already provided.
        For example, set shallow clone if branch != (master|main).
        """
        # If user doesn't specify 'git_clone_depth', pick default
        if "git_clone_depth" not in self.config:
            is_main_branch = self.config.get("branch", "") not in ["master", "main"]
            self.config["git_clone_depth"] = "--depth 1" if is_main_branch else ""

        version = self.config.get("version", "")
        branch = self.config.get("branch", "")
        if version:
            # e.g. treat version as git branch or tag
            if branch:
                print(f"'version'='{version}' is explicitly set and overrides 'branch'='{branch}' (this might be desired)")
            self.config["branch"] = version

        # For convenience, build a 'clone_command' string
        self.config["clone_command"] = (
            "git clone {git_clone_depth} -b {branch} {url} {source_path}"
            .format(**self.config)
        )

    def fetch(self):
        repo_url = self.config.get("url", "")
        source_path = self.config.get("source_path", "")
        clone_command = self.config.get("clone_command", "")

        if not repo_url:
            raise ValueError(
                "[GitFetcher] 'url' is missing in config. Current config: {}".format(self.config)
            )

        # If already cloned or source_path is not empty, skip
        if os.path.exists(source_path) and os.path.isdir(source_path) and os.listdir(source_path):
            # The directory exists and is not empty. Do nothing.
            return

        # Ensure the parent directories exist
        run('mkdir -p "{}"'.format(source_path))

        # Execute the clone
        run(clone_command)

    def use_common_dirs_scheme(self):
        """Function sets common directory scheme."""
        if 'app_path' in self.config:
            # The directory, where binary is installed
            if not "install_path" in self.config:
                self.config["install_path"] = "{app_path}/{app_name}-{branch}".format(**self.config)


class TarballFetcher(IFetcher):
    """
    A fetcher that downloads a tarball from a URL, extracts it,
    and places contents in `source_path`.
    """

    def preconfigure(self):
        # Optionally refine or default something, e.g. local temp name
        if "tar_temp_name" not in self.config:
            self.config["tar_temp_name"] = "/tmp/temp.tar.gz"

    def fetch(self):
        file_url = self.config.get("file_url", "")
        app_path = self.config.get("app_path", "")
        source_path = os.path.join(app_path, "src")  # or use source_path from config

        if not file_url:
            raise ValueError("[TarballFetcher] 'file_url' not specified in config.")

        # Create the source_path
        run('mkdir -p "{}"'.format(source_path))

        download_cmd = f"wget {file_url} -O {self.config['tar_temp_name']}"
        run(download_cmd)

        extract_cmd = f"tar zxvf {self.config['tar_temp_name']} -C {source_path} --strip-components=1"
        run(extract_cmd)


class FileSystemFetcher(IFetcher):
    """
    A fetcher that simply uses a local directory as 'source_path'.
    Optionally can copy it, or do nothing if the user wants to build in-place.
    """

    def fetch(self):
        # The user might store it in config["path"]
        # or config["source_path"]. Decide which to rely on:
        path = self.config.get("path", "")
        source_path = self.config.get("source_path", "")

        if not path:
            raise ValueError("[FileSystemFetcher] No 'path' provided in config.")
        if not os.path.isdir(path):
            raise ValueError(f"[FileSystemFetcher] Provided 'path' is not a directory: {path}")

        # A typical usage: just copy or do nothing. We'll do a naive copy example:
        # If user actually wants an in-place usage, they can skip copying.
        # For demonstration, let's copy to source_path if different:
        if source_path and source_path != path:
            run(f'mkdir -p "{source_path}"')
            # A naive copy with rsync (example):
            copy_cmd = f'rsync -a "{path}/" "{source_path}/"'
            run(copy_cmd)
        else:
            # If user sets them the same, we do nothing.
            pass


def make_fetcher(config: Dict[str, Any]) -> IFetcher:
    """
    Factory that picks the fetcher based on config['fetch'] or tries to autodetect
    from 'fetch' value if it is a URL or local path.
    """
    fetch_val = config.get("fetch", "")
    if not fetch_val:
        # No fetch step
        return None

    # If user explicitly says "git", "tarball", or "filesystem"
    if fetch_val in ("git", "tarball", "filesystem"):
        if fetch_val == "git":
            return GitFetcher(config)
        elif fetch_val == "tarball":
            return TarballFetcher(config)
        else:
            return FileSystemFetcher(config)

    # Otherwise, do an autodetect:
    if fetch_val.endswith(".git"):
        return GitFetcher(config)
    elif fetch_val.endswith(".tar.gz"):
        return TarballFetcher(config)
    else:
        # assume local filesystem
        return FileSystemFetcher(config)



# tests/test_cli_cmd_pwd.py
import os
import pytest
from click.testing import CliRunner
import click
from unittest.mock import patch, MagicMock

from edpm.cli.pwd import pwd_command
from edpm.engine.api import EdpmApi

# Create a test CLI that includes only the 'pwd' command
@click.group()
@click.pass_context
def cli(ctx):
    pass

cli.add_command(pwd_command, "pwd")


@pytest.fixture
def runner():
    """Provides a Click CliRunner for capturing output."""
    return CliRunner()


@pytest.fixture
def mock_api():
    """Create a mock API with controlled behavior for testing"""
    api = MagicMock(spec=EdpmApi)
    api.top_dir = "/mock/top/dir"

    # Mock plan with a test package
    mock_plan = MagicMock()
    mock_plan.find_package.return_value = True  # Any package exists in plan
    mock_plan.find_package.side_effect = lambda name: name in ["test-pkg", "existing-pkg"]
    api.plan = mock_plan

    # Mock lock with test package installed
    mock_lock = MagicMock()
    mock_lock.is_installed.return_value = False  # Default: not installed
    mock_lock.is_installed.side_effect = lambda name: name in ["test-pkg", "existing-pkg"]

    # Set up mock package data
    test_pkg_data = {
        "install_path": "/mock/top/dir/test-pkg/install",
        "source_path": "/mock/top/dir/test-pkg/source",
        "build_path": "/mock/top/dir/test-pkg/build",
        "owned": True
    }

    existing_pkg_data = {
        "install_path": "/external/path/existing-pkg",
        "owned": False
    }

    def get_installed_package(name):
        if name == "test-pkg":
            return test_pkg_data
        elif name == "existing-pkg":
            return existing_pkg_data
        return None

    mock_lock.get_installed_package.side_effect = get_installed_package
    api.lock = mock_lock

    return api


def test_pwd_no_args(runner, mock_api):
    """Test 'edpm pwd' with no arguments - should show top dir"""
    result = runner.invoke(cli, ["pwd"], obj=mock_api)
    assert result.exit_code == 0
    assert "/mock/top/dir" in result.output


def test_pwd_with_package(runner, mock_api):
    """Test 'edpm pwd test-pkg' - should show all paths"""
    result = runner.invoke(cli, ["pwd", "test-pkg"], obj=mock_api)
    assert result.exit_code == 0
    assert "Install path:" in result.output
    assert "/mock/top/dir/test-pkg/install" in result.output
    assert "Sources:" in result.output
    assert "/mock/top/dir/test-pkg/source" in result.output
    assert "Build dir:" in result.output
    assert "/mock/top/dir/test-pkg/build" in result.output
    assert "edpm 'owned' base path:" in result.output


def test_pwd_with_existing_package(runner, mock_api):
    """Test with an existing (not owned) package"""
    result = runner.invoke(cli, ["pwd", "existing-pkg"], obj=mock_api)
    assert result.exit_code == 0
    assert "Install path:" in result.output
    assert "/external/path/existing-pkg" in result.output
    assert "edpm 'owned' base path:" not in result.output  # Should not show for non-owned


def test_pwd_install_only(runner, mock_api):
    """Test 'edpm pwd test-pkg --install' - should only show install path"""
    result = runner.invoke(cli, ["pwd", "test-pkg", "--install"], obj=mock_api)
    assert result.exit_code == 0
    assert result.output.strip() == "/mock/top/dir/test-pkg/install"


def test_pwd_source_only(runner, mock_api):
    """Test 'edpm pwd test-pkg --source' - should only show source path"""
    result = runner.invoke(cli, ["pwd", "test-pkg", "--source"], obj=mock_api)
    assert result.exit_code == 0
    assert result.output.strip() == "/mock/top/dir/test-pkg/source"


def test_pwd_build_only(runner, mock_api):
    """Test 'edpm pwd test-pkg --build' - should only show build path"""
    result = runner.invoke(cli, ["pwd", "test-pkg", "--build"], obj=mock_api)
    assert result.exit_code == 0
    assert result.output.strip() == "/mock/top/dir/test-pkg/build"


def test_pwd_package_not_found(runner, mock_api):
    """Test with a package that doesn't exist in the plan"""
    result = runner.invoke(cli, ["pwd", "unknown-pkg"], obj=mock_api)
    assert result.exit_code != 0
    assert "Error: Package 'unknown-pkg' not found in the plan" in result.output


def test_pwd_package_not_installed(runner, mock_api):
    """Test with a package that exists in plan but is not installed"""
    # Temporarily modify the mock behavior
    original_side_effect = mock_api.lock.is_installed.side_effect
    mock_api.lock.is_installed.side_effect = lambda name: False

    result = runner.invoke(cli, ["pwd", "test-pkg"], obj=mock_api)

    # Restore original behavior
    mock_api.lock.is_installed.side_effect = original_side_effect

    assert result.exit_code != 0
    assert "Error: Package 'test-pkg' is not installed" in result.output


import os
import sys
import click
from typing import Dict, List, Set, Tuple, Optional

from edpm.engine.api import EdpmApi
from edpm.engine.output import markup_print as mprint

@click.command("req")
@click.argument('os_name', nargs=1, metavar='<os-name>')
@click.argument('package_names', nargs=-1, metavar='<package-names>')
@click.option('--optional', 'print_mode', flag_value='optional', help="Print optional packages")
@click.option('--required', 'print_mode', flag_value='required', help="Print required packages")
@click.option('--all', 'print_mode', flag_value='all', help="Print all packages (ready for package manager install)")
@click.option('--all-titles', 'print_mode', flag_value='all_titles', help="Print all packages (human readable)", default=True)
def req_command(ctx, os_name, package_names, print_mode):
    """Get list of system requirements for specified packages.

    Example:
        edpm req ubuntu root
        edpm req centos root geant4
    """
    assert isinstance(ctx.obj, EdpmApi)
    api = ctx.obj
    api.load_all()

    # OS mapping - define supported OS types and aliases
    os_aliases = {
        "ubuntu": "ubuntu",
        "debian": "ubuntu",  # Use ubuntu packages for debian
        "centos": "centos",
        "rhel": "centos",    # Use centos packages for RHEL
        "fedora": "centos",  # Use centos packages for fedora
    }

    # Check if OS is supported
    if os_name.lower() not in os_aliases:
        mprint('<red><b>ERROR</b></red>: OS name "{}" is unknown\nKnown OS names are:', os_name)
        for name in sorted(os_aliases.keys()):
            mprint('   {}', name)
        sys.exit(1)

    normalized_os = os_aliases[os_name.lower()]

    # Get requirements
    required_pkgs, optional_pkgs = gather_requirements(api, normalized_os, package_names)

    # Display results according to print_mode
    if print_mode == 'optional':
        mprint(" ".join(optional_pkgs))
    elif print_mode == 'required':
        mprint(" ".join(required_pkgs))
    elif print_mode == 'all':
        mprint(" ".join(required_pkgs + optional_pkgs))
    else:  # all_titles
        mprint("<blue><b>REQUIRED</b></blue>:")
        for pkg in sorted(required_pkgs):
            mprint("  {}", pkg)

        if optional_pkgs:
            mprint("\n<blue><b>OPTIONAL</b></blue>:")
            for pkg in sorted(optional_pkgs):
                mprint("  {}", pkg)


def gather_requirements(api: EdpmApi, normalized_os: str, package_names: Tuple[str]) -> Tuple[List[str], List[str]]:
    """Gather requirements from packages and their recipes."""
    required_pkgs: Set[str] = set()
    optional_pkgs: Set[str] = set()

    # If no packages specified, get requirements for all packages in the plan
    if not package_names:
        package_names = [p.name for p in api.plan.packages()]

    # Get global requirements if they exist
    global_config = api.plan.global_config()
    global_require = global_config.get("require", {})

    # Extract from global require section
    os_reqs = global_require.get(normalized_os, {})
    if isinstance(os_reqs, list):
        required_pkgs.update(os_reqs)
    elif isinstance(os_reqs, dict):
        required_pkgs.update(os_reqs.get("required", []))
        optional_pkgs.update(os_reqs.get("optional", []))

    # Get package-specific requirements
    for pkg_name in package_names:
        # Get the package from the plan
        pkg = api.plan.find_package(pkg_name)
        if not pkg:
            mprint('<yellow>Warning:</yellow> Package "{}" not found in the plan', pkg_name)
            continue

        # Check for require section in package config
        if hasattr(pkg, 'config') and isinstance(pkg.config, dict):
            pkg_require = pkg.config.get("require", {})
            os_reqs = pkg_require.get(normalized_os, {})
            if isinstance(os_reqs, list):
                required_pkgs.update(os_reqs)
            elif isinstance(os_reqs, dict):
                required_pkgs.update(os_reqs.get("required", []))
                optional_pkgs.update(os_reqs.get("optional", []))

        # Also check if there's a recipe with os_dependencies
        recipe_name = pkg_name
        # Try to get recipe class from recipe manager
        recipe_cls = api.recipe_manager.recipes_by_name.get(recipe_name)
        if recipe_cls:
            if hasattr(recipe_cls, 'os_dependencies') and isinstance(recipe_cls.os_dependencies, dict):
                # Extract from recipe's os_dependencies
                required = recipe_cls.os_dependencies.get('required', {}).get(normalized_os, "")
                optional = recipe_cls.os_dependencies.get('optional', {}).get(normalized_os, "")

                if required:
                    if isinstance(required, str):
                        required_pkgs.update(required.split())
                    elif isinstance(required, list):
                        required_pkgs.update(required)

                if optional:
                    if isinstance(optional, str):
                        optional_pkgs.update(optional.split())
                    elif isinstance(optional, list):
                        optional_pkgs.update(optional)

    # Make sure no package is both required and optional
    optional_pkgs = optional_pkgs - required_pkgs

    return list(required_pkgs), list(optional_pkgs)


import os
from typing import Any, Dict, List, Optional
from ruamel.yaml import YAML
from edpm.engine.generators.steps import GeneratorStep

yaml_rt = YAML(typ='rt')  # round-trip mode

def expand_placeholders(text: str, placeholders: Dict[str, str]) -> str:
    for k, v in placeholders.items():
        text = text.replace(f'${k}', v)
    return text

class EnvironmentBlock:
    """
    Holds a list of environment instructions (set/prepend/append).
    """
    def __init__(self, data: List[Any]):
        # data is an array of objects like:  [{set: {...}}, {prepend: {...}}]
        self.data = data or []

    def parse(self, placeholders: Optional[Dict[str, str]] = None) -> List[GeneratorStep]:
        """
        Convert environment instructions into GeneratorStep objects.
        placeholders can be used to expand e.g. "$install_dir"
        """
        from edpm.engine.generators.steps import EnvSet, EnvPrepend, EnvAppend
        results: List[GeneratorStep] = []

        if placeholders is None:
            placeholders = {}

        for item in self.data:
            if not isinstance(item, dict):
                # skip or raise error
                continue

            # item is like {"prepend": {"PATH": "$install_dir/bin"}}
            # or {"set": {...}}
            for action_key, kv_dict in item.items():
                if not isinstance(kv_dict, dict):
                    continue
                for var_name, raw_val in kv_dict.items():
                    expanded_val = expand_placeholders(str(raw_val), placeholders)
                    if action_key == "set":
                        results.append(EnvSet(var_name, expanded_val))
                    elif action_key == "prepend":
                        results.append(EnvPrepend(var_name, expanded_val))
                    elif action_key == "append":
                        results.append(EnvAppend(var_name, expanded_val))
                    else:
                        pass  # unknown action

        return results

class PlanPackage:
    """
    Represents one dependency from the plan.

    This can be:
      1) A "baked in" name (like "root" or "geant4"), possibly with @version
      2) A custom dictionary with 'fetch', 'make', environment, etc.

    `config` is just a dict, no separate ConfigBlock anymore.
    `env_block_obj` is an EnvironmentBlock for environment instructions if present.
    """

    def __init__(self, name: str, config_data: Dict[str, Any], env_data: List[Any], is_baked_in: bool = False):
        self._name = name
        self._is_baked_in = is_baked_in
        # config is a raw dict of fields (fetch, make, branch, etc.)
        self.config = config_data
        # environment instructions for just this package
        self.env_block_obj = EnvironmentBlock(env_data)

    @property
    def name(self) -> str:
        return self._name

    @property
    def is_baked_in(self) -> bool:
        return self._is_baked_in

    def env_block(self) -> EnvironmentBlock:
        return self.env_block_obj

class PlanFile:
    def __init__(self, raw_data: Dict[str, Any]):
        self.data = raw_data or {}

        # Ensure 'global' sub-dict
        if "global" not in self.data:
            self.data["global"] = {}

        # Ensure 'packages' is a list
        if "packages" not in self.data or not isinstance(self.data["packages"], list):
            self.data["packages"] = []

        # Ensure there's a 'config' in global
        if "config" not in self.data["global"]:
            self.data["global"]["config"] = {
                "build_threads": 4,
                "cxx_standard": 17
            }

        # Ensure there's an 'environment' in global
        if "environment" not in self.data["global"]:
            self.data["global"]["environment"] = []

    @classmethod
    def load(cls, filename: str) -> "PlanFile":
        if not os.path.isfile(filename):
            raise FileNotFoundError(f"Plan file not found: {filename}")
        yaml_rt.preserve_quotes = True
        with open(filename, "r", encoding="utf-8") as f:
            raw_data = yaml_rt.load(f) or {}
        return cls(raw_data)

    def save(self, filename: str):
        with open(filename, "w", encoding="utf-8") as f:
            yaml_rt.dump(self.data, f)

    def global_config(self) -> Dict[str, Any]:
        """
        Return the raw dict for global config (instead of a ConfigBlock).
        """
        return self.data["global"]["config"]

    def get_global_env_actions(self) -> List[GeneratorStep]:
        block = EnvironmentBlock(self.data["global"]["environment"])
        return block.parse()

    def packages(self) -> List[PlanPackage]:
        """
        Parse the 'packages' array into a list[PlanPackage].
        Each item can be:
          - A string: "root" or "geant4@v11.03"
          - A dict: { "mydep": { fetch:..., environment:..., etc. } }
        """
        pkg_list = self.data["packages"]
        result: List[PlanPackage] = []

        for item in pkg_list:
            if isinstance(item, str):
                # e.g. "root" or "geant4@v11.03"
                pkg_name = item
                version_part = ""
                if '@' in pkg_name:
                    parts = pkg_name.split('@', 1)
                    pkg_name = parts[0]
                    version_part = parts[1]

                config_data = {}
                if version_part:
                    config_data["version"] = version_part

                p = PlanPackage(
                    name=pkg_name,
                    config_data=config_data,
                    env_data=[],
                    is_baked_in=True
                )
                result.append(p)

            elif isinstance(item, dict):
                # e.g. { "mylib": { fetch:..., environment: [...], etc. } }
                if len(item) != 1:
                    raise ValueError(
                        f"Malformed dependency entry. Must have exactly one top-level key.\n"
                        f"Invalid entry: {item}"
                    )
                dep_name, dep_config = next(iter(item.items()))
                if not isinstance(dep_config, dict):
                    raise ValueError(
                        f"Invalid config for dependency '{dep_name}'. Must be a dictionary.\n"
                        f"Got: {type(dep_config)}"
                    )
                env_data = dep_config.get("environment", [])
                # shallow copy so we can remove environment key
                tmp_config = dict(dep_config)
                tmp_config.pop("environment", None)

                p = PlanPackage(
                    name=dep_name,
                    config_data=tmp_config,
                    env_data=env_data,
                    is_baked_in=False
                )
                result.append(p)
            else:
                raise ValueError(f"Invalid package entry: {item} (type {type(item)})")

        return result

    def has_package(self, name: str) -> bool:
        """
        True if a package with name 'name' is in the plan.
        """
        return any(p.name == name for p in self.packages())

    def find_package(self, name: str) -> Optional[PlanPackage]:
        for p in self.packages():
            if p.name == name:
                return p
        return None

    def add_package(self, new_entry: Any):
        """
        Append a new package item (string or dict) to self.data["packages"].
        Must remain a list. Then the user can call save() to persist.
        """
        if not isinstance(self.data["packages"], list):
            self.data["packages"] = []

        self.data["packages"].append(new_entry)



"""
This file provides information of how to build and configure EIC Reconstruction (eicrecon) framework:
https://gitlab.com/eic/eicrecon

"""
import os

from edpm.engine.generators.steps import EnvSet, EnvPrepend, CmakePrefixPath
from edpm.engine.composed_recipe import ComposedRecipe


class EicreconRecipe(ComposedRecipe):
    """
    Installs eicrecon (EIC Reconstruction Framework) from Git + CMake.
    """
    def __init__(self, config):
        self.default_config = {
            'fetch': 'git',
            'make': 'cmake',
            'url': 'https://github.com/eic/eicrecon.git',
            'branch': 'main'
        }
        super().__init__(name='eicrecon', config=config)

    @staticmethod
    def gen_env(data):
        path = data['install_path']
        lib_path = os.path.join(path, 'lib')
        lib64_path = os.path.join(path, 'lib64')

        yield EnvSet('eicrecon_HOME', path)

        yield EnvPrepend('JANA_PLUGIN_PATH', os.path.join(path, 'lib', 'EICrecon', 'plugins'))
        yield EnvPrepend('PATH', os.path.join(path, 'bin'))

        if os.path.isdir(lib64_path):
            yield EnvPrepend('LD_LIBRARY_PATH', lib64_path)
        yield EnvPrepend('LD_LIBRARY_PATH', lib_path)




# edpm/cli/env.py

import os
import click
from edpm.engine.api import EdpmApi


@click.group("env")
@click.pass_context
def env_group(ctx):
    """
    Manages environment and integration files.

    Subcommands:
      edpm env bash        -> print the bash environment script
      edpm env csh         -> print the csh environment script
      edpm env cmake       -> print the EDPM toolchain .cmake text
      edpm env cmake-prof  -> print the CMakePresets.json text
      edpm env save        -> saves environment scripts & CMake files
    """
    if ctx.obj is None:
        ctx.obj = EdpmApi()
    ctx.obj.load_all()


@env_group.command("bash")
@click.pass_context
def env_bash(ctx):
    """
    Print the pure EDPM Bash environment script (no merging).
    """
    api = ctx.obj
    env_gen = api.create_environment_generator()
    edpm_content = env_gen.build_env_text(shell="bash")
    click.echo(edpm_content)


@env_group.command("csh")
@click.pass_context
def env_csh(ctx):
    """
    Print the pure EDPM C Shell environment script (no merging).
    """
    api = ctx.obj
    env_gen = api.create_environment_generator()
    edpm_content = env_gen.build_env_text(shell="csh")
    click.echo(edpm_content)


@env_group.command("cmake")
@click.pass_context
def env_cmake(ctx):
    """
    Print the EDPM-generated toolchain config (no merging).
    """
    api = ctx.obj
    cm_gen = api.create_cmake_generator()
    text = cm_gen.build_toolchain_text()
    click.echo(text)


@env_group.command("cmake-prof")
@click.pass_context
def env_cmake_prof(ctx):
    """
    Print the EDPM-generated CMakePresets JSON (no merging).
    """
    api = ctx.obj
    cm_gen = api.create_cmake_generator()
    text = cm_gen.build_presets_json()
    click.echo(text)


@env_group.command("save")
@click.pass_context
def env_save(ctx):
    """
    Saves environment scripts & CMake files.

    The Plan file can have, in `global.config`:
      env_bash_in,    env_bash_out,
      env_csh_in,     env_csh_out,
      cmake_toolchain_in,  cmake_toolchain_out,
      cmake_presets_in,    cmake_presets_out

    If *either* the _in or _out is blank (""), we skip.
    If a key is omitted (None), we fallback to planDir + default name for _out
    and we do *not* do any merging for _in.

    For example:
      global:
        config:
          env_bash_in: /home/user/myBaseEnv.sh
          env_bash_out: /home/user/mergedEnv.sh
          cmake_presets_in: /home/user/originalPresets.json
          cmake_presets_out: /home/user/finalPresets.json
          # etc...
    """

    api = ctx.obj

    api.save_generator_scripts()
